import asyncio
import copy
import inspect
import logging
from threading import RLock
from typing import Dict, Optional, Any, Callable, Tuple

import sleekxmpp
import sleekxmpp.exceptions
from sleekxmpp.plugins.xep_0009.binding import fault2xml, xml2fault, xml2py, py2xml

from pyobs.modules import Module
from pyobs.comm.exceptions import *
from pyobs.utils.parallel import Future

log = logging.getLogger(__name__)


class RPC(object):
    """RPC wrapper around XEP0009."""

    def __init__(self, client: sleekxmpp.ClientXMPP, loop, handler: Optional[Module] = None):
        """Create a new RPC wrapper.

        Args:
            client: XMPP client tu use for communication.
            handler: pyobs module that handles function calls.
        """

        # store
        self._client = client
        self._lock = RLock()
        self._futures: Dict[str, Future] = {}
        self._handler = handler
        self._methods: Dict[str, Tuple[Callable[[], Any], inspect.Signature]] = {}
        self._loop = loop

        # set up callbacks
        client.add_event_handler("jabber_rpc_method_call", self._on_jabber_rpc_method_call)
        client.add_event_handler("jabber_rpc_method_timeout", self._on_jabber_rpc_method_timeout)
        client.add_event_handler("jabber_rpc_method_response", self._on_jabber_rpc_method_response)
        client.add_event_handler("jabber_rpc_method_fault", self._on_jabber_rpc_method_fault)
        client.add_event_handler("jabber_rpc_error", self._on_jabber_rpc_error)

        # register handler
        self.set_handler(handler)

    def set_handler(self, handler: Optional[Module] = None) -> None:
        """Set the handler for remote procedure calls to this client.

        Args:
            handler: Handler object.
        """

        # store handler
        self._handler = handler

        # update methods
        self._methods = copy.copy(handler.methods) if handler else {}

    async def call(self, target_jid: str, method: str, signature: inspect.Signature, *args: Any) -> Any:
        """Call a method on a remote host.

        Args:
            target_jid: Target JID to call method on.
            method: Name of method to call.
            signature: Method signature.
            *args: Parameters for method.

        Returns:
            Future for response.
        """

        # create the method call
        iq = self._client.plugin["xep_0009"].make_iq_method_call(target_jid, method, py2xml(*args))

        # create a future for this
        pid = iq["id"]
        future = Future(signature=signature)
        self._futures[pid] = future

        # send request
        try:
            iq.send(block=False)
        except sleekxmpp.exceptions.IqError:
            raise InvocationException()
        except sleekxmpp.exceptions.IqTimeout:
            raise TimeoutException()

        # don't wait for response, just return future
        return await future

    def _on_jabber_rpc_method_call(self, iq: Any) -> None:
        """React on remote method call.

        Args:
            iq: Received XMPP message.
        """

        # get method and parameters
        iq.enable("rpc_query")
        params = xml2py(iq["rpc_query"]["method_call"]["params"])
        pmethod = iq["rpc_query"]["method_call"]["method_name"]

        # no handler?
        if self._handler is None:
            raise ValueError("No handler specified.")

        # get method
        with self._lock:
            try:
                method, signature = self._methods[pmethod]

            except KeyError:
                log.error("No handler available for %s!", pmethod)
                self._client.plugin["xep_0009"].item_not_found(iq).send()
                return

            # bind parameters
            ba = signature.bind(*params)
            ba.apply_defaults()

        # call method
        asyncio.run_coroutine_threadsafe(self._call_remote(iq, pmethod, method, ba.arguments), self._loop)

    async def _call_remote(self, iq: Any, method_name: str, method: Callable, arguments: dict) -> None:
        """React on remote method call.

        Args:
            iq: Received XMPP message.
        """

        try:
            # do we have a timeout?
            if hasattr(method, "timeout"):
                timeout = await getattr(method, "timeout")(self._handler, **arguments)
                if timeout:
                    # yes, send it!
                    response = self._client.plugin["xep_0009_timeout"].make_iq_method_timeout(
                        iq["id"], iq["from"], int(timeout)
                    )
                    response.send()

            # call method
            return_value = await self._handler.execute(method_name, **arguments, sender=iq["from"].user)
            return_value = () if return_value is None else (return_value,)

            # send response
            self._client.plugin["xep_0009"].make_iq_method_response(iq["id"], iq["from"], py2xml(*return_value)).send()

        except InvocationException as ie:
            # could not invoke method
            fault = {"code": 500, "string": ie.get_message()}
            self._client.plugin["xep_0009"].send_fault(iq, fault2xml(fault))

        except Exception as e:
            # something else went wrong
            log.warning("Error during call to %s: %s", method_name, str(e), exc_info=True)

            # send response
            fault = dict()
            fault["code"] = 500
            fault["string"] = str(e)
            self._client.plugin["xep_0009"].send_fault(iq, fault2xml(fault))

    def _on_jabber_rpc_method_response(self, iq: Any) -> None:
        """Received a response for a method call.

        Args:
            iq: Received XMPP message.
        """
        asyncio.run_coroutine_threadsafe(self._method_response(iq), self._loop)

    async def _method_response(self, iq: Any) -> None:
        """Received a response for a method call.

        Args:
            iq: Received XMPP message.
        """

        # get message
        iq.enable("rpc_query")
        try:
            args = xml2py(iq["rpc_query"]["method_response"]["params"])
        except ValueError:
            log.error("Could not parse method response: %s", iq)
            return

        # get future
        pid = iq["id"]
        with self._lock:
            if pid not in self._futures:
                return
            future = self._futures[pid]
            del self._futures[pid]

        # set result of future
        if not future.done():
            if len(args) > 0:
                future.set_result(args[0])
            else:
                future.set_result(None)

    def _on_jabber_rpc_method_timeout(self, iq: Any) -> None:
        """Method call timed out.

        Args:
            iq: Received XMPP message.
        """
        iq.enable("rpc_query")
        timeout = iq["rpc_query"]["method_timeout"]["timeout"]
        pid = iq["id"]
        self._futures[pid].set_timeout(timeout)

    def _on_jabber_rpc_method_fault(self, iq: Any) -> None:
        """Communication to host failed.

        Args:
            iq: Received XMPP message.
        """

        # get message
        iq.enable("rpc_query")
        fault = xml2fault(iq["rpc_query"]["method_response"]["fault"])

        # get future
        pid = iq["id"]
        with self._lock:
            future = self._futures[pid]
            del self._futures[pid]

        # set error
        future.set_exception(InvocationException(fault["string"]))

    def _on_jabber_rpc_error(self, iq: Any) -> None:
        """Method invocation failes.

        Args:
            iq: Received XMPP message.
        """

        # get message
        pmethod = self._client.plugin["xep_0009"].extract_method(iq["rpc_query"])
        condition = iq["error"]["condition"]

        # get future
        pid = iq["id"]
        with self._lock:
            callback = self._futures[pid]
            del self._futures[pid]

        # set error
        e = {
            "item-not-found": RemoteException("No remote handler available for %s at %s!" % (pmethod, iq["from"])),
            "forbidden": AuthorizationException(
                "Forbidden to invoke remote handler for %s at %s!" % (pmethod, iq["from"])
            ),
            "undefined-condition": RemoteException(
                "An unexpected problem occured trying to invoke %s at %s!" % (pmethod, iq["from"])
            ),
            "service-unavailable": RemoteException("The service at %s is unavailable." % iq["from"]),
            "remote-server-not-found": RemoteException("Could not find remote server for %s." % iq["from"]),
        }[condition]
        if e is None:
            RemoteException("An unexpected exception occurred at %s!" % iq["from"])
        callback.set_exception(e)


__all__ = ["RPC"]



import logging
import threading
import time
from datetime import datetime
from typing import Any
import numpy as np
from astropy.io import fits

from pyobs.modules.camera.basespectrograph import BaseSpectrograph
from pyobs.utils.enums import ExposureStatus


log = logging.getLogger(__name__)


class DummySpectrograph(BaseSpectrograph):
    """A dummy spectrograph for testing."""

    __module__ = "pyobs.modules.camera"

    def __init__(self, **kwargs: Any):
        """Creates a new dummy cammera.

        Args:
            readout_time: Readout time in seconds.
            sim: Dictionary with config for image simulator.
        """
        BaseSpectrograph.__init__(self, **kwargs)

    async def _expose(self, abort_event: threading.Event) -> fits.HDUList:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image and, if present, a filename.

        Raises:
            ValueError: If exposure was not successful.
        """

        # do exposure
        log.info("Starting exposure...")
        exposure_time = 1.0
        date_obs = datetime.utcnow()
        self._exposing = True
        steps = 10
        for i in range(steps):
            if abort_event.is_set() or not self._exposing:
                self._exposing = False
                await self._change_exposure_status(ExposureStatus.IDLE)
                raise ValueError("Exposure was aborted.")
            time.sleep(exposure_time / steps)
        self._exposing = False

        # readout
        await self._change_exposure_status(ExposureStatus.READOUT)
        time.sleep(1)

        # get data
        x = np.arange(5000, 8000, 1)
        y = np.sin(x)

        # get spectrum
        hdu = fits.PrimaryHDU(y)

        # add headers
        hdu.header["DATE-OBS"] = date_obs.strftime("%Y-%m-%dT%H:%M:%S.%f")
        hdu.header["CRVAL1"] = 5000
        hdu.header["CDELT1"] = 1

        # finished
        log.info("Exposure finished.")
        return fits.HDUList([hdu])

    def _abort_exposure(self) -> None:
        """Abort the running exposure. Should be implemented by derived class.

        Returns:
            Success or not.
        """
        self._exposing = False

    async def get_exposure_progress(self, **kwargs: Any) -> float:
        """Returns the progress of the current exposure in percent.

        Returns:
            Progress of the current exposure in percent.
        """
        return 1.0


__all__ = ["DummySpectrograph"]



import logging
import threading
import slixmpp
import slixmpp.xmlstream
from typing import Any

from pyobs.comm.slixmpp.xep_0009.rpc import XEP_0009
from pyobs.comm.slixmpp.xep_0009_timeout import XEP_0009_timeout


log = logging.getLogger(__name__)


class XmppClient(slixmpp.ClientXMPP):  # type: ignore
    """XMPP client for pyobs."""

    def __init__(self, jid: str, password: str):
        """Create a new XMPP client.

        An XmppClient handles the actual XMPP communication for the XmppComm module.

        Args:
            jid: Connect to the XMPP server with the given JID.
            password: Password to use for connection.
        """

        slixmpp.ClientXMPP.__init__(self, jid, password)

        # stuff
        self._connect_event = threading.Event()
        self._logs_node = "logs"
        self._auth_event = threading.Event()
        self._auth_success = False

        # auto-accept invitations
        self.auto_authorize = True

        # register plugins
        self.register_plugin("xep_0009", module=XEP_0009)  # RPC
        self.register_plugin("xep_0009_timeout", module=XEP_0009_timeout)  # RPC timeout
        self.register_plugin("xep_0030")  # Service Discovery
        self.register_plugin("xep_0045")  # Multi-User Chat
        self.register_plugin("xep_0060")  # PubSub
        self.register_plugin("xep_0115")  # Entity Capabilities
        self.register_plugin("xep_0163")  # Personal Eventing Protocol
        self.register_plugin("xep_0199")  # XMPP Ping

        # enable keep alive pings
        self["xep_0199"].enable_keepalive()

        # handle session_start and message events
        self.add_event_handler("session_start", self.session_start)
        self.add_event_handler("auth_success", lambda ev: self._auth(True))
        self.add_event_handler("failed_auth", lambda ev: self._auth(False))

    def wait_connect(self) -> bool:
        """Wait for client to connect.

        Returns:
            Success or not.
        """

        # wait for auth and check
        self._auth_event.wait()
        if not self._auth_success:
            log.error("Invalid credentials for connecting to XMPP server.")
            return False

        # wait for final connect
        self._connect_event.wait()

        # connected
        return True

    def session_start(self, event: Any) -> None:
        """Session start event.

        Args:
            event: The event sent at session start.
        """
        log.info("Connected to server.")

        # send presence and get roster
        self.send_presence()
        self.get_roster()

        # send connected event
        self._connect_event.set()

    def _auth(self, success: bool) -> None:
        """Called after authentification.

        Args:
            success: Whether or not the authentification was successful.
        """
        # store and fire
        self._auth_success = success
        self._auth_event.set()


__all__ = ["XmppClient"]



from typing import Optional, Any, Dict
from typing_extensions import TypedDict

from pyobs.events.event import Event
from pyobs.utils.enums import ExposureStatus


DataType = TypedDict("DataType", {"last": Optional[str], "current": str})


class ExposureStatusChangedEvent(Event):
    """Event to be sent, when the exposure status of a device changes."""

    __module__ = "pyobs.events"

    def __init__(self, current: ExposureStatus, last: Optional[ExposureStatus] = None, **kwargs: Any):
        Event.__init__(self)
        self.data: DataType = {
            "last": last.value if last is not None else None,
            "current": current.value,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        # get current
        if "current" not in d or not isinstance(d["current"], str):
            raise ValueError("Invalid type for current.")
        current: ExposureStatus = ExposureStatus(d["current"])

        # get last
        last: Optional[ExposureStatus] = None
        if "last" in d and isinstance(d["last"], str):
            last = ExposureStatus(d["last"])

        # return object
        return ExposureStatusChangedEvent(current=current, last=last)

    @property
    def last(self) -> Optional[ExposureStatus]:
        return ExposureStatus(self.data["last"]) if self.data["last"] is not None else None

    @property
    def current(self) -> ExposureStatus:
        return ExposureStatus(self.data["current"])


__all__ = ["ExposureStatusChangedEvent"]



import asyncio
import logging
import platform
import signal
import warnings
import threading
from io import StringIO
from logging.handlers import TimedRotatingFileHandler
from typing import Optional, Any, Dict

import yaml

from pyobs.object import get_object
from pyobs.modules import Module
from pyobs.utils.config import pre_process_yaml

# just init logger with something here, will be overwritten in __init__
log = logging.getLogger(__name__)


class Application:
    """Class for initializing and shutting down a pyobs process."""

    def __init__(self, config: str, log_file: Optional[str] = None, log_level: str = "info", **kwargs: Any):
        """Initializes a pyobs application.

        Args:
            config: Name of config file.
            log_file: Name of log file, if any.
            log_level: Logging level.
            log_rotate: Whether to rotate the log files.
        """

        # get config name without path and extension
        self._config = config

        # formatter for logging, and list of logging handlers
        formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(filename)s:%(lineno)d %(message)s")
        handlers = []

        # create stdout logging handler
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(formatter)
        handlers.append(stream_handler)

        # create file logging handler, if log file is given
        if log_file is not None:
            # in Windows, append a FileHandler, otherwise we use a WatchedFileHandler, which works well with logrotate
            if platform.system() == "Windows":
                file_handler = logging.FileHandler(log_file)
            else:
                file_handler = logging.handlers.WatchedFileHandler(log_file)

            # add log file handler
            file_handler.setFormatter(formatter)
            handlers.append(file_handler)

        # basic setup
        logging.basicConfig(handlers=handlers, level=logging.getLevelName(log_level.upper()))
        logging.captureWarnings(True)
        warnings.simplefilter("always", DeprecationWarning)

        # disable tornado logger
        logging.getLogger("tornado.access").disabled = True

        # set pyobs logger
        global log
        log = logging.getLogger(__name__)

        # hack threading to set thread names on OS level
        self._hack_threading()

        # load config
        log.info("Loading configuration from {0:s}...".format(self._config))
        with StringIO(pre_process_yaml(self._config)) as f:
            cfg: Dict[str, Any] = yaml.safe_load(f)

        # create module and open it
        log.info("Creating module...")
        self._module = get_object(cfg, Module)

    def run(self) -> None:
        """Run app."""

        # signals
        loop = asyncio.get_event_loop()
        for sig in (signal.SIGTERM, signal.SIGINT):
            loop.add_signal_handler(sig, self._signal_handler, sig)

        # run main task forever
        loop.create_task(self._main())
        loop.run_forever()

        # main finished, cancel all tasks
        tasks = asyncio.all_tasks(loop=loop)
        for t in tasks:
            t.cancel()
        group = asyncio.gather(*tasks, return_exceptions=True)
        loop.run_until_complete(group)

        # finished
        loop.close()

    def _signal_handler(self, sig) -> None:
        """React to signals and quit module."""

        # stop loop
        loop = asyncio.get_running_loop()
        loop.stop()

        # reset signal handlers
        log.info(f"Got signal: {sig!s}, shutting down.")
        loop.remove_signal_handler(signal.SIGTERM)
        loop.add_signal_handler(signal.SIGINT, lambda: None)

    async def _main(self) -> None:
        """Actually run the application."""

        # everything in a try/except/finally, so that we can shut down gracefully
        try:
            # open module
            log.info("Opening module...")
            await self._module.open()
            log.info("Started successfully.")

            # run module
            await self._module.main()

        except:
            # some exception was thrown
            log.exception("Something went wrong.")

        finally:
            # shutting down
            log.info("Shutting down...")

            # close module
            if self._module is not None:
                log.info("Closing module...")
                await self._module.close()

            # finished
            log.info("Finished shutting down.")

    def _hack_threading(self) -> None:
        """Bad hack to set thread name on OS level."""
        try:
            import prctl

            def set_thread_name(name: str) -> None:
                prctl.set_name(name)

            def _thread_name_hack(this: Any) -> None:
                set_thread_name(this.name)
                threading.Thread.__bootstrap_original__(this)  # type: ignore

            threading.Thread.__bootstrap_original__ = threading.Thread._bootstrap  # type: ignore
            threading.Thread._bootstrap = _thread_name_hack  # type: ignore

        except ImportError:
            logger = logging.getLogger("pyobs")
            logger.warning("prctl module is not installed. You will not be able to see thread names")

            def set_thread_name(name: str) -> None:
                pass


class GuiApplication(Application):
    """Derived Application class that uses a Qt GUI. Allows for graceful shutdown in Windows."""

    def __init__(self, **kwargs: Any):
        """Create a new GUI application."""
        Application.__init__(self, **kwargs)

        # import Qt stuff
        from PyQt5.QtWidgets import QApplication
        from pyobs.utils.modulegui import ModuleGui
        import sys

        # create Qt app and window
        self._qapp = QApplication(sys.argv)
        self._window = ModuleGui()

        # show window
        self._window.show()

    def _run(self) -> None:
        """Run qt application."""
        self._qapp.exec()


__all__ = ["Application", "GuiApplication"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IAbortable(Interface, metaclass=ABCMeta):
    """The module has an abortable action."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def abort(self, **kwargs: Any) -> None:
        """Abort current actions."""
        ...


__all__ = ["IAbortable"]



from abc import ABCMeta, abstractmethod

from pyobs.images import Image
from pyobs.images.processor import ImageProcessor


class Astrometry(ImageProcessor, metaclass=ABCMeta):
    """Base class for astrometry processors"""

    __module__ = "pyobs.images.processors.astrometry"

    @abstractmethod
    async def __call__(self, image: Image) -> Image:
        """Find astrometric solution on given image.

        Args:
            image: Image to analyse.

        Returns:
            Processed image.
        """
        ...


__all__ = ["Astrometry"]



import asyncio
import glob
import logging
from datetime import datetime
from typing import Tuple, NamedTuple, Dict, Any, Optional, TYPE_CHECKING

from pyobs.interfaces import IWindow, IBinning, ICooling, IGain
from pyobs.modules.camera.basecamera import BaseCamera
from pyobs.images import Image
from pyobs.utils.enums import ExposureStatus
from pyobs.utils import exceptions as exc

if TYPE_CHECKING:
    from pyobs.utils.simulation import SimWorld


log = logging.getLogger(__name__)


class CoolingStatus(NamedTuple):
    enabled: bool = True
    set_point: float = -10.0
    power: float = 80.0
    temperatures: Dict[str, float] = {"CCD": 0.0, "Back": 3.14}


class DummyCamera(BaseCamera, IWindow, IBinning, ICooling, IGain):
    """A dummy camera for testing."""

    __module__ = "pyobs.modules.camera"

    def __init__(
        self,
        readout_time: float = 2,
        sim: Optional[Dict[str, Any]] = None,
        world: Optional["SimWorld"] = None,
        **kwargs: Any,
    ):
        """Creates a new dummy cammera.

        Args:
            readout_time: Readout time in seconds.
            sim: Dictionary with config for image simulator.
        """
        BaseCamera.__init__(self, **kwargs)

        # add thread func
        self.add_background_task(self._cooling_thread, True)

        # store
        self._readout_time = readout_time
        self._sim = sim if sim is not None else {}
        if "images" not in self._sim:
            self._sim["images"] = None

        # simulated world
        from pyobs.utils.simulation import SimCamera, SimWorld

        self._world = self.get_object(world, SimWorld) if world is not None else self.add_child_object(SimWorld)
        self._camera: SimCamera = self._world.camera

        # init camera
        self._cooling = CoolingStatus()
        self._exposing = True
        self._gain = 10.0

        # simulator
        self._sim_images = sorted(glob.glob(self._sim["images"])) if self._sim["images"] else None

    async def _cooling_thread(self) -> None:
        while True:
            # adjust temperature
            temps = self._cooling.temperatures
            temps["CCD"] -= (self._cooling.temperatures["CCD"] - self._cooling.set_point) * 0.05

            # cooling power
            power = (60.0 - self._cooling.temperatures["CCD"]) / 70.0 * 100.0

            # create new object
            self._cooling = CoolingStatus(
                enabled=self._cooling.enabled, set_point=self._cooling.set_point, power=power, temperatures=temps
            )

            # sleep for 1 second
            await asyncio.sleep(1)

    async def get_full_frame(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns full size of CCD.

        Returns:
            Tuple with left, top, width, and height set.
        """
        return self._camera.full_frame

    def _get_image(self, exp_time: float, open_shutter: bool) -> Image:
        """Actually get (i.e. simulate) the image."""

        # random image or pre-defined?
        if self._sim_images:
            filename = self._sim_images.pop(0)
            self._sim_images.append(filename)
            return Image.from_file(filename)

        else:
            image = self._camera.get_image(exp_time, open_shutter)
            return image

    async def _expose(self, exposure_time: float, open_shutter: bool, abort_event: asyncio.Event) -> Image:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            exposure_time: The requested exposure time in seconds.
            open_shutter: Whether or not to open the shutter.
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image.

        Raises:
            GrabImageError: If exposure was not successful.
        """

        # start exposure
        log.info("Starting exposure with {0:s} shutter...".format("open" if open_shutter else "closed"))
        date_obs = datetime.utcnow()
        self._exposing = True

        # request image
        loop = asyncio.get_running_loop()
        hdu_future = loop.run_in_executor(None, self._get_image, exposure_time, open_shutter)

        # wait a little
        steps = 10
        for i in range(steps):
            if abort_event.is_set() or not self._exposing:
                self._exposing = False
                await self._change_exposure_status(ExposureStatus.IDLE)
                raise InterruptedError("Exposure was aborted.")
            await asyncio.sleep(exposure_time / steps)
        self._exposing = False

        # readout
        await self._change_exposure_status(ExposureStatus.READOUT)
        await asyncio.sleep(self._readout_time)

        # get image
        image = await hdu_future

        # add headers
        image.header["EXPTIME"] = exposure_time
        image.header["DATE-OBS"] = date_obs.strftime("%Y-%m-%dT%H:%M:%S.%f")
        image.header["XBINNING"] = image.header["DET-BIN1"] = (self._camera.binning[0], "Binning factor used on X axis")
        image.header["YBINNING"] = image.header["DET-BIN2"] = (self._camera.binning[1], "Binning factor used on Y axis")
        image.header["XORGSUBF"] = (self._camera.window[0], "Subframe origin on X axis")
        image.header["YORGSUBF"] = (self._camera.window[1], "Subframe origin on Y axis")

        # biassec/trimsec
        self.set_biassec_trimsec(image.header, *self._camera.full_frame)

        # finished
        log.info("Exposure finished.")
        return image

    async def _abort_exposure(self) -> None:
        """Abort the running exposure. Should be implemented by derived class.

        Returns:
            Success or not.
        """
        self._exposing = False

    async def get_window(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns the camera window.

        Returns:
            Tuple with left, top, width, and height set.
        """
        return self._camera.window

    async def set_window(self, left: int, top: int, width: int, height: int, **kwargs: Any) -> None:
        """Set the camera window.

        Args:
            left: X offset of window.
            top: Y offset of window.
            width: Width of window.
            height: Height of window.

        Raises:
            ValueError: If binning could not be set.
        """
        log.info("Set window to %dx%d at %d,%d.", width, height, top, left)
        self._camera.window = (left, top, width, height)

    async def get_binning(self, **kwargs: Any) -> Tuple[int, int]:
        """Returns the camera binning.

        Returns:
            Tuple with x and y.
        """
        return self._camera.binning

    async def set_binning(self, x: int, y: int, **kwargs: Any) -> None:
        """Set the camera binning.

        Args:
            x: X binning.
            y: Y binning.

        Raises:
            ValueError: If binning could not be set.
        """
        log.info("Set binning to %dx%d.", x, y)
        self._camera.binning = (x, y)

    async def set_cooling(self, enabled: bool, setpoint: float, **kwargs: Any) -> None:
        """Enables/disables cooling and sets setpoint.

        Args:
            enabled: Enable or disable cooling.
            setpoint: Setpoint in celsius for the cooling.

        Raises:
            ValueError: If cooling could not be set.
        """

        # log
        if enabled:
            log.info("Enabling cooling with a setpoint of %.2f°C.", setpoint)
        else:
            log.info("Disabling cooling.")

        # set
        self._cooling = CoolingStatus(
            enabled=enabled, set_point=setpoint, power=self._cooling.power, temperatures=self._cooling.temperatures
        )

    async def get_cooling_status(self, **kwargs: Any) -> Tuple[bool, float, float]:
        """Returns the current status for the cooling.

        Returns:
            (tuple): Tuple containing:
                Enabled:  Whether the cooling is enabled
                SetPoint: Setpoint for the cooling in celsius.
                Power:    Current cooling power in percent or None.
        """
        return self._cooling.enabled, self._cooling.set_point, self._cooling.power

    async def get_temperatures(self, **kwargs: Any) -> Dict[str, float]:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """
        return self._cooling.temperatures

    def _set_config_readout_time(self, readout_time: float) -> None:
        """Set readout time."""
        self._readout_time = readout_time

    def _get_config_readout_time(self) -> float:
        """Returns readout time."""
        return self._readout_time

    async def set_gain(self, gain: float, **kwargs: Any) -> None:
        """Set the camera gain.

        Args:
            gain: New camera gain.

        Raises:
            ValueError: If gain could not be set.
        """
        log.info("Setting gain to %.2f...", gain)
        self._gain = gain

    async def get_gain(self, **kwargs: Any) -> float:
        """Returns the camera binning.

        Returns:
            Current gain.
        """
        return self._gain


__all__ = ["DummyCamera"]



from __future__ import annotations

import logging
from typing import Dict, Union, Tuple, Optional, List
from astroplan import Observer
import operator

from .exptimeeval import ExpTimeEval
from .priorities import SkyflatPriorities
from ..time import Time


log = logging.getLogger(__name__)


class SchedulerItem:
    """A single item in the flat scheduler"""

    def __init__(self, start: float, end: float, filter_name: str, binning: Tuple[int, int], priority: float):
        """Initializes a new scheduler item

        Args:
            start: Start time in seconds
            end: End time in seconds
            filter_name: Name of filter
            binning: Used binning
        """
        self.start = start
        self.end = end
        self.filter_name = filter_name
        self.binning = binning
        self.priority = priority

    def __repr__(self):
        """Nice string representation for item"""
        return "%d - %d (%s %dx%d): %.2f" % (
            self.start,
            self.end,
            self.filter_name,
            self.binning[0],
            self.binning[1],
            self.priority,
        )


class Scheduler:
    """Scheduler for taking flat fields"""

    __module__ = "pyobs.utils.skyflats"

    def __init__(
        self,
        functions: Union[str, Dict[str, Union[str, Dict[str, str]]]],
        priorities: SkyflatPriorities,
        observer: Observer,
        min_exptime: float = 0.5,
        max_exptime: float = 5,
        timespan: float = 7200,
        filter_change: float = 30,
        count: int = 20,
        readout: Optional[Dict[str, float]] = None,
    ):
        """Initializes a new scheduler for taking flat fields

        Args:
            functions: Flat field functions
            priorities: Class handling priorities
            observer: Observer to use
            min_exptime: Minimum exposure time for flats
            max_exptime: Maximum exposure time for flats
            timespan: Timespan from now that should be scheduled [s]
            filter_change: Time required for filter change [s]
            count: Number of flats to schedule
            readout: Dictionary with readout times (in sec) per binning (as BxB).
        """
        self._eval = ExpTimeEval(observer, functions)
        self._observer = observer
        self._priorities = priorities
        self._min_exptime = min_exptime
        self._max_exptime = max_exptime
        self._schedules: List[SchedulerItem] = []
        self._current = 0
        self._timespan = timespan
        self._filter_change = filter_change
        self._count = count
        self._readout = {} if readout is None else readout

    async def __call__(self, time: Time) -> None:
        """Calculate schedule starting at given time

        Args:
            time: Time to start schedule at
        """

        # init evaluator
        self._eval.init(time)

        # sort filters by priority in a List of (filter, binning, priority) tuples
        tmp = [(k, v) for k, v in (await self._priorities()).items()]
        priorities = sorted(tmp, key=operator.itemgetter(1), reverse=True)

        # place them
        schedules: List[SchedulerItem] = []
        for (filter_name, binning), priority in priorities:
            # find possible time
            self._find_slot(schedules, filter_name, binning, priority)

        # sort by start time
        self._schedules = sorted(schedules, key=lambda x: x.start)

    def __iter__(self) -> Scheduler:
        """Iterator for scheduler items"""
        self._current = 0
        return self

    def __next__(self) -> SchedulerItem:
        """Iterate over scheduler items"""
        if self._current < len(self._schedules):
            item = self._schedules[self._current]
            self._current += 1
            return item
        else:
            raise StopIteration

    def __getitem__(self, item):
        """Return schedule item."""
        return self._schedules[item]

    def _find_slot(
        self, schedules: List[SchedulerItem], filter_name: str, binning: Tuple[int, int], priority: float
    ) -> None:
        """Find a possible slot for a given filter/binning in the given schedule

        Args:
            schedules: List of existing schedules
            filter_name: Name of filter
            binning: Used binning
        """

        # get readout time
        sbin = "%dx%d" % binning
        readout = self._readout[sbin] if sbin in self._readout else 0.0

        # find first possible start time
        time = 0
        while time < self._timespan:
            # get exposure time
            exp_time_start = self._eval.exp_time(time, binning=binning, filter_name=filter_name)

            # are we in allowed limit?
            if self._min_exptime <= exp_time_start <= self._max_exptime:
                # seems to fit, get duration
                duration = self._eval.duration(
                    self._count, start_time=time, readout=readout, binning=binning, filter_name=filter_name
                )

                # add time for filter change
                duration += self._filter_change

                # get exp time at end
                exp_time_end = self._eval.exp_time(time + duration, binning=binning, filter_name=filter_name)

                # still in limits?
                if self._min_exptime <= exp_time_end <= self._max_exptime:
                    # check for overlap with existing schedule
                    if not self._overlaps(schedules, time, time + duration):
                        # add schedule and quit
                        schedules.append(SchedulerItem(time, time + duration, filter_name, binning, priority))
                        return

            # next step
            time += 10

        # being here means we didn't find any
        return

    def _overlaps(self, schedules: List[SchedulerItem], start: float, end: float) -> bool:
        """Checks, whether a new scheduler item would overlap an existing item

        Args:
            schedules: List of existing scheduler items
            start: Start time of new item
            end: End time of new item

        Returns:
            Whether it overlaps
        """

        # loop all scheduler items
        item: SchedulerItem
        for item in schedules:
            # does it overlap?
            if (start < item.end and end > item.start) or (item.start < end and item.end > start):
                return True

        # no overlap found
        return False


__all__ = ["SchedulerItem", "Scheduler"]



"""
Utilities
TODO: write doc
"""
__title__ = "Utility modules"

from .autonomouswarning import AutonomousWarning
from .fluentlogger import FluentLogger
from .httpfilecache import HttpFileCache
from .kiosk import Kiosk
from .telegram import Telegram
from .trigger import Trigger



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IExposureTime(Interface, metaclass=ABCMeta):
    """The camera supports exposure times, to be used together with :class:`~pyobs.interfaces.ICamera`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_exposure_time(self, exposure_time: float, **kwargs: Any) -> None:
        """Set the exposure time in seconds.

        Args:
            exposure_time: Exposure time in seconds.

        Raises:
            ValueError: If exposure time could not be set.
        """
        ...

    @abstractmethod
    async def get_exposure_time(self, **kwargs: Any) -> float:
        """Returns the exposure time in seconds.

        Returns:
            Exposure time in seconds.
        """
        ...

    @abstractmethod
    async def get_exposure_time_left(self, **kwargs: Any) -> float:
        """Returns the remaining exposure time on the current exposure in seconds.

        Returns:
            Remaining exposure time in seconds.
        """
        ...


__all__ = ["IExposureTime"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class ICalibrate(Interface, metaclass=ABCMeta):
    """The module can calibrate a device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def calibrate(self, **kwargs: Any) -> None:
        """Calibrate the device."""
        ...


__all__ = ["ICalibrate"]



import asyncio
import logging
from functools import partial
from typing import Union, List, Optional, Any, Dict
import numpy as np
import astropy.units as u

from pyobs.mixins.pipeline import PipelineMixin
from pyobs.object import Object
from pyobs.utils.archive import Archive
from pyobs.utils.enums import ImageType
from pyobs.images import Image, ImageProcessor
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class Pipeline(Object, PipelineMixin):
    """Pipeline based on the astropy package ccdproc."""

    def __init__(self, steps: List[Union[Dict[str, Any], ImageProcessor]], **kwargs: Any):
        """Pipeline for science images.

        Args:
            steps: List of pipeline steps to perform.
        """
        Object.__init__(self, **kwargs)
        PipelineMixin.__init__(self, steps)

    @staticmethod
    def _combine_calib_images(
        images: List[Image], bias: Optional[Image] = None, normalize: bool = False, method: str = "average"
    ) -> Image:
        """Combine a list of given images.

        Args:
            images: List of images to combine.
            bias: If given, subtract from images before combining them.
            normalize: If True, images are normalized to median of 1 before and after combining them.
            method: Method for combining images.
        """
        import ccdproc

        # get CCDData objects
        data = [image.to_ccddata() for image in images]

        # subtract bias?
        if bias is not None:
            bias_data = bias.to_ccddata()
            data = [ccdproc.subtract_bias(d, bias_data) for d in data]

        # normalize?
        if normalize:
            data = [d.divide(np.median(d.data), handle_meta="first_found") for d in data]

        # combine image
        combined = ccdproc.combine(
            data,
            method=method,
            sigma_clip=True,
            sigma_clip_low_thresh=5,
            sigma_clip_high_thresh=5,
            mem_limit=350e6,
            unit="adu",
            combine_uncertainty_function=np.ma.std,
        )

        # normalize?
        if normalize:
            combined = combined.divide(np.median(combined.data), handle_meta="first_found")

        # to Image and copy header
        image = Image.from_ccddata(combined)

        # add history
        for i, src in enumerate(images, 1):
            basename = src.header["FNAME"].replace(".fits.fz", "").replace(".fits", "")
            image.header["L1AVG%03d" % i] = (basename, "Image used for average")
        image.header["RLEVEL"] = (1, "Reduction level")

        # finished
        return image

    async def _combine_calib_images_async(self, images: List[Image], **kwargs: Any) -> Image:
        return await asyncio.get_running_loop().run_in_executor(
            None, partial(self._combine_calib_images, images, **kwargs)
        )

    async def create_master_bias(self, images: List[Image]) -> Image:
        """Create master bias frame.

        Args:
            images: List of raw bias frames.

        Returns:
            Master bias frame.
        """
        return await self._combine_calib_images_async(images)

    async def create_master_dark(self, images: List[Image], bias: Image) -> Image:
        """Create master dark frame.

        Args:
            images: List of raw dark frames.
            bias: Bias frame to subtract from images.

        Returns:
            Master dark frame.
        """
        return await self._combine_calib_images_async(images, bias=bias)

    async def create_master_flat(self, images: List[Image], bias: Image) -> Image:
        """Create master flat frame.

        Args:
            images: List of raw flat frames.
            bias: Bias frame to subtract from images.

        Returns:
            Master flat frame.
        """
        return await self._combine_calib_images_async(images, bias=bias, normalize=True, method="median")

    async def calibrate(self, image: Image) -> Image:
        """Calibrate a single science frame.

        Args:
            image: Image to calibrate.

        Returns:
            Calibrated image.
        """

        # copy image
        calibrated = image.copy()

        # run pipeline
        return await self.run_pipeline(calibrated)

    @staticmethod
    async def find_master(
        archive: Archive,
        image_type: ImageType,
        time: Time,
        instrument: str,
        binning: str,
        filter_name: Optional[str] = None,
        max_days: Optional[float] = 30.0,
    ) -> Optional[Image]:
        """Find and download master calibration frame.

        Args:
            archive: Image archive.
            image_type: Image type.
            time: Time to search at.
            instrument: Instrument to use.
            binning: Used binning.
            filter_name: Used filter.
            max_days: Maximum number of days from DATE-OBS to find frames.

        Returns:
            FrameInfo for master calibration frame or None.
        """

        # find reduced frames from +- N days
        log.info(
            "Searching for %s %s master calibration frames%s from instrument %s.",
            binning,
            image_type.value,
            "" if filter_name is None else " in " + filter_name,
            instrument,
        )
        infos = await archive.list_frames(
            start=None if max_days is None else time - max_days * u.day,
            end=None if max_days is None else time + max_days * u.day,
            instrument=instrument,
            image_type=image_type,
            binning=binning,
            filter_name=filter_name,
            rlevel=1,
        )

        # found none?
        if len(infos) == 0:
            log.warning("Could not find any matching %s calibration frames.", image_type.value)
            return None

        # sort by diff to time and take first
        s = sorted(infos, key=lambda i: abs((i.dateobs - time).sec))
        info = s[0]
        log.info("Found %s frame %s.", image_type.name, info.filename)

        # download it
        data = await archive.download_frames([info])
        return data[0]


__all__ = ["Pipeline"]



from abc import ABCMeta, abstractmethod

from pyobs.images import Image
from pyobs.images.processor import ImageProcessor


class Photometry(ImageProcessor, metaclass=ABCMeta):
    """Base class for photometry processors."""

    __module__ = "pyobs.images.processors.photometry"

    @abstractmethod
    async def __call__(self, image: Image) -> Image:
        """Do aperture photometry on given image.

        Args:
            image: Image to do aperture photometry on.

        Returns:
            Image with attached catalog.
        """
        ...


__all__ = ["Photometry"]



from __future__ import annotations
from typing import Optional, Dict, Any
from typing_extensions import TypedDict

from pyobs.events.event import Event
from pyobs.utils.enums import ImageType


DataType = TypedDict("DataType", {"filename": str, "image_type": Optional[str], "raw": Optional[str]})


class NewImageEvent(Event):
    """Event to be sent on a new image."""

    __module__ = "pyobs.events"

    def __init__(self, filename: str, image_type: Optional[ImageType] = None, raw: Optional[str] = None, **kwargs: Any):
        """Initializes new NewImageEvent.

        Args:
            filename: Name of new image file.
            image_type: Type of image.
            raw: Only for reduced images, references raw frame.
        """
        Event.__init__(self)
        self.data: DataType = {
            "filename": filename,
            "image_type": image_type.value if image_type is not None else None,
            "raw": raw,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        # get filename
        if "filename" not in d or not isinstance(d["filename"], str):
            raise ValueError("Invalid type for filename.")
        filename: str = d["filename"]

        # get image type
        image_type: Optional[ImageType] = None
        if "image_type" in d and isinstance(d["image_type"], str):
            image_type = ImageType(d["image_type"])

        # get raw
        raw: Optional[str] = None
        if "raw" in d and isinstance(d["raw"], str):
            raw = d["raw"]

        # return object
        return NewImageEvent(filename, image_type, raw)

    @property
    def filename(self) -> str:
        return self.data["filename"]

    @property
    def image_type(self) -> Optional[ImageType]:
        return ImageType(self.data["image_type"]) if "image_type" in self.data else None

    @property
    def raw(self) -> Optional[str]:
        return self.data["raw"]

    @property
    def is_reduced(self) -> bool:
        return self.raw is not None


__all__ = ["NewImageEvent"]



import asyncio
import logging
import time
from typing import Optional, Union, Any

from pyobs.interfaces import IWeather
from pyobs.modules import Module
from pyobs.events import BadWeatherEvent, GoodWeatherEvent, Event
from pyobs.interfaces import IMotion
from pyobs.mixins import MotionStatusMixin
from pyobs.utils.enums import MotionStatus
from pyobs.utils.parallel import event_wait

log = logging.getLogger(__name__)


class WeatherAwareMixin:
    """Mixin for IMotion devices that should park(), when weather gets bad."""

    __module__ = "pyobs.mixins"

    def __init__(self, weather: Optional[Union[str, IWeather]] = None, **kwargs: Any):
        self.__weather = weather
        self.__is_weather_good: Optional[bool] = None
        self.__last_park_attempt: Optional[float] = None
        self.__setting_bad_weather = False
        this = self
        if isinstance(self, Module):
            self.add_background_task(this.__weather_check, True)
        else:
            raise ValueError("This is not a module.")

    async def open(self) -> None:
        """Open mixin."""
        # subscribe to events
        this = self
        if isinstance(self, Module) and self.comm is not None:
            await self.comm.register_event(BadWeatherEvent, this.__on_bad_weather)
            await self.comm.register_event(GoodWeatherEvent, this._on_good_weather)

    async def __on_bad_weather(self, event: Event, sender: str) -> bool:
        """Abort exposure if a bad weather event occurs.

        Args:
            event: The bad weather event.
            sender: Who sent it.
        """

        # check
        if not isinstance(event, BadWeatherEvent):
            raise ValueError("Wrong event type.")

        # park
        asyncio.create_task(self.__set_bad_weather())
        return True

    async def __set_bad_weather(self) -> None:
        # already setting?
        if self.__setting_bad_weather:
            return
        self.__setting_bad_weather = True

        # did weather change?
        if self.__is_weather_good is True:
            # weather is bad now
            self.__is_weather_good = False

            # do we need to park?
            if isinstance(self, MotionStatusMixin) and isinstance(self, IMotion):
                motion_status = await self.get_motion_status()
                if motion_status == MotionStatus.ERROR:
                    log.error("Telescope is in error mode, cannot park.")
                elif motion_status != MotionStatus.PARKED:
                    log.warning("Weather is bad, shutting down.")
                    await self.park()
            else:
                log.error("This is not a MotionStatusMixin/IMotion.")

        # finished
        self.__setting_bad_weather = False

    async def _on_good_weather(self, event: Event, sender: str) -> bool:
        """Change status of weather.

        Args:
            event: The good weather event.
            sender: Who sent it.
        """

        # check
        if not isinstance(event, GoodWeatherEvent):
            raise ValueError("Wrong event type.")

        # weather is good
        self.__is_weather_good = True
        return True

    async def __weather_check(self) -> None:
        """Thread for continuously checking for good weather"""

        # module?
        this = self
        if isinstance(self, Module):
            module = self

            # run until closing
            while True:
                # got a weather module?
                if this.__weather is None:
                    # weather is always good
                    this.__is_weather_good = True

                else:
                    try:
                        # get proxy
                        weather: IWeather = await module.proxy(this.__weather, IWeather)

                        # get good status
                        this.__is_weather_good = await weather.is_weather_good()

                    except:
                        # could either not connect or weather is not good
                        this.__is_weather_good = False

                # if not good, park now
                if this.__is_weather_good is False:
                    asyncio.create_task(this.__set_bad_weather())

                # sleep a little
                await asyncio.sleep(10)

        else:
            # not a module
            raise ValueError("This is not a module.")

    def is_weather_good(self) -> bool:
        return self.__is_weather_good is True


__all__ = ["WeatherAwareMixin"]



import copy
import inspect
import logging
from typing import Dict, Optional, Any, Callable, Tuple
import slixmpp.exceptions

from pyobs.modules import Module
from pyobs.utils.parallel import Future
from pyobs.comm.slixmpp.xep_0009.binding import fault2xml, xml2fault, xml2py, py2xml
import pyobs.utils.exceptions as exc

log = logging.getLogger(__name__)


class RPC(object):
    """RPC wrapper around XEP0009."""

    def __init__(self, client: slixmpp.ClientXMPP, handler: Optional[Module] = None):
        """Create a new RPC wrapper.

        Args:
            client: XMPP client tu use for communication.
            handler: pyobs module that handles function calls.
        """

        # store
        self._client = client
        self._futures: Dict[str, Future] = {}
        self._handler = handler
        self._methods: Dict[str, Tuple[Callable[[], Any], inspect.Signature]] = {}

        # set up callbacks
        client.add_event_handler("jabber_rpc_method_call", self._on_jabber_rpc_method_call)
        client.add_event_handler("jabber_rpc_method_timeout", self._on_jabber_rpc_method_timeout)
        client.add_event_handler("jabber_rpc_method_response", self._on_jabber_rpc_method_response)
        client.add_event_handler("jabber_rpc_method_fault", self._on_jabber_rpc_method_fault)
        client.add_event_handler("jabber_rpc_error", self._on_jabber_rpc_error)

        # register handler
        self.set_handler(handler)

    def set_handler(self, handler: Optional[Module] = None) -> None:
        """Set the handler for remote procedure calls to this client.

        Args:
            handler: Handler object.
        """

        # store handler
        self._handler = handler

        # update methods
        self._methods = copy.copy(handler.methods) if handler else {}

    async def call(self, target_jid: str, method: str, signature: inspect.Signature, *args: Any) -> Any:
        """Call a method on a remote host.

        Args:
            target_jid: Target JID to call method on.
            method: Name of method to call.
            signature: Method signature.
            *args: Parameters for method.

        Returns:
            Future for response.
        """

        # create the method call
        iq = self._client.plugin["xep_0009"].make_iq_method_call(target_jid, method, py2xml(*args))

        # create a future for this
        pid = iq["id"]
        future = Future(signature=signature)
        self._futures[pid] = future

        # send request
        await iq.send()

        # don't wait for response, just return future
        return await future

    async def _on_jabber_rpc_method_call(self, iq: Any) -> None:
        """React on remote method call.

        Args:
            iq: Received XMPP message.
        """

        # get method and parameters
        iq.enable("rpc_query")
        params = xml2py(iq["rpc_query"]["method_call"]["params"])
        pmethod = iq["rpc_query"]["method_call"]["method_name"]

        try:
            # no handler?
            if self._handler is None:
                return
                # raise ValueError('No handler specified.')

            # get method
            try:
                method, signature = self._methods[pmethod]
            except KeyError:
                log.error("No handler available for %s!", pmethod)
                self._client.plugin["xep_0009"].item_not_found(iq).send()
                return

            # bind parameters
            ba = signature.bind(*params)
            ba.apply_defaults()

            # do we have a timeout?
            if hasattr(method, "timeout"):
                timeout = await getattr(method, "timeout")(self._handler, **ba.arguments)
                if timeout:
                    # yes, send it!
                    response = self._client.plugin["xep_0009_timeout"].make_iq_method_timeout(
                        iq["id"], iq["from"], int(timeout)
                    )
                    response.send()

            # call method
            return_value = await self._handler.execute(pmethod, *params, sender=iq["from"].user)
            return_value = () if return_value is None else (return_value,)

            # send response
            self._client.plugin["xep_0009"].make_iq_method_response(iq["id"], iq["from"], py2xml(*return_value)).send()

        # except InvocationException as ie:
        #    # could not invoke method
        #    self._client.plugin["xep_0009"].send_fault(iq, fault2xml(500, ie.get_message()))

        except exc.PyObsError as e:
            # something else went wrong, but only log if not a ModuleError
            if not isinstance(e, exc.ModuleError):
                log.error("Error during call to %s: %s", pmethod, str(e), exc_info=True)

            # send response
            self._client.plugin["xep_0009"].send_fault(iq, fault2xml(500, str(e)))

    async def _on_jabber_rpc_method_response(self, iq: Any) -> None:
        """Received a response for a method call.

        Args:
            iq: Received XMPP message.
        """

        # get message
        iq.enable("rpc_query")
        try:
            args = xml2py(iq["rpc_query"]["method_response"]["params"])
        except ValueError:
            log.error("Could not parse method response: %s", iq)
            return

        # get future
        pid = iq["id"]
        if pid not in self._futures:
            return
        future = self._futures[pid]
        del self._futures[pid]

        # set result of future, if it's not set already (probably with an exception)
        if not future.done():
            if len(args) > 0:
                future.set_result(args[0])
            else:
                future.set_result(None)

    async def _on_jabber_rpc_method_timeout(self, iq: Any) -> None:
        """Method call timed out.

        Args:
            iq: Received XMPP message.
        """
        iq.enable("rpc_query")
        timeout = iq["rpc_query"]["method_timeout"]["timeout"]
        pid = iq["id"]
        self._futures[pid].set_timeout(timeout)

    async def _on_jabber_rpc_method_fault(self, iq: Any) -> None:
        """Communication to host failed.

        Args:
            iq: Received XMPP message.
        """

        # get message
        iq.enable("rpc_query")
        fault = xml2fault(iq["rpc_query"]["method_response"]["fault"])

        # get future
        jid = iq["id"]
        future = self._futures[jid]
        del self._futures[jid]

        # get exception and error
        s: str = fault["string"]
        exception_name = s[1 : s.index(">")]
        exception_message = s[s.index(">") + 1 :].strip()

        # get class of exception
        exception_class = getattr(exc, exception_name)

        # and instantiate it
        if issubclass(exception_class, exc.RemoteError):
            exception = exception_class(message=exception_message, module=jid.node)
        else:
            exception = exception_class(message=exception_message)

        # sender
        sender = iq["from"].node

        # set error
        if not future.done():
            future.set_exception(exc.InvocationError(module=sender, exception=exception))

    async def _on_jabber_rpc_error(self, iq: Any) -> None:
        """Method invocation failes.

        Args:
            iq: Received XMPP message.
        """

        # get message
        pmethod = self._client.plugin["xep_0009"].extract_method(iq["rpc_query"])
        condition = iq["error"]["condition"]

        # get future
        pid = iq["id"]
        callback = self._futures[pid]
        del self._futures[pid]

        # sender
        sender = iq["from"].node

        # set error
        e = {
            "item-not-found": exc.RemoteError(sender, f"No remote handler available for {pmethod} at {iq['from']}'!"),
            "forbidden": exc.RemoteError(sender, f"Forbidden to invoke remote handler for {pmethod} at {iq['from']}!"),
            "undefined-condition": exc.RemoteError(
                sender, f"An unexpected problem occured trying to invoke {pmethod} at {iq['from']}!"
            ),
            "service-unavailable": exc.RemoteError(sender, f"The service at {iq['from']} is unavailable."),
            "remote-server-not-found": exc.RemoteError(sender, f"Could not find remote server for {iq['from']}."),
        }[condition]
        if e is None:
            exc.RemoteError(sender, f"An unexpected exception occurred at {iq['from']}!")
        callback.set_exception(e)


__all__ = ["RPC"]



from __future__ import annotations
from typing import Dict, Any
from typing_extensions import TypedDict

from pyobs.events.event import Event


DataType = TypedDict("DataType", {"filename": str})


class NewSpectrumEvent(Event):
    """Event to be sent on a new image."""

    __module__ = "pyobs.events"

    def __init__(self, filename: str, **kwargs: Any):
        """Initializes new NewSpectrumEvent.

        Args:
            filename: Name of new image file.
        """
        Event.__init__(self)
        self.data: DataType = {"filename": filename}

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        # get filename
        if "filename" not in d or not isinstance(d["filename"], str):
            raise ValueError("Invalid type for filename.")
        filename: str = d["filename"]

        # return object
        return NewSpectrumEvent(filename)

    @property
    def filename(self) -> str:
        return self.data["filename"]


__all__ = ["NewSpectrumEvent"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class IPointingRaDec(Interface, metaclass=ABCMeta):
    """The module can move to RA/Dec coordinates, usually combined with :class:`~pyobs.interfaces.ITelescope`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def move_radec(self, ra: float, dec: float, **kwargs: Any) -> None:
        """Starts tracking on given coordinates.

        Args:
            ra: RA in deg to track.
            dec: Dec in deg to track.

        Raises:
            MoveError: If device could not be moved.
        """
        ...

    @abstractmethod
    async def get_radec(self, **kwargs: Any) -> Tuple[float, float]:
        """Returns current RA and Dec.

        Returns:
            Tuple of current RA and Dec in degrees.
        """
        ...


__all__ = ["IPointingRaDec"]



import io
from typing import List, Dict, Any, Optional, cast
import aiohttp
import urllib.parse
import logging

from pyobs.utils.time import Time
from pyobs.images import Image
from pyobs.utils.archive import Archive, FrameInfo
from pyobs.utils.enums import ImageType

log = logging.getLogger(__name__)


class PyobsArchiveFrameInfo(FrameInfo):
    """Frame info for pyobs archive."""

    def __init__(self, info: Dict[str, str]):
        FrameInfo.__init__(self)
        self.info = info
        self.id = self.info["id"]
        self.filename = self.info["basename"]
        self.dateobs = Time(self.info["DATE_OBS"])
        self.filter_name = self.info["FILTER"]
        self.binning = int(self.info["binning"][0])
        self.url = self.info["url"]


class PyobsArchive(Archive):
    """Connector class to running pyobs-archive instance."""

    __module__ = "pyobs.utils.archive"

    def __init__(self, url: str, token: str, proxies: Optional[Dict[str, str]] = None, **kwargs: Any):
        self._url = url
        self._headers = {"Authorization": "Token " + token}
        self._proxies = proxies

    async def list_options(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> Dict[str, List[Any]]:
        # build URL
        url = urllib.parse.urljoin(self._url, "frames/aggregate/")

        # and params
        params = self._build_query(
            start, end, night, site, telescope, instrument, image_type, binning, filter_name, rlevel
        )

        # do request
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params, headers=self._headers, timeout=10) as response:
                if response.status != 200:
                    raise ValueError("Could not query frames: %s" % str(await response.text()))

                # create frames and return them
                return await response.json()

    async def list_frames(
        self,
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> List[FrameInfo]:
        # build URL
        url = urllib.parse.urljoin(self._url, "frames/")

        # and params
        params = self._build_query(
            start, end, night, site, telescope, instrument, image_type, binning, filter_name, rlevel
        )

        # init list
        frames = []

        # set offset and limit
        params["offset"] = 0
        params["limit"] = 1000

        # open session
        async with aiohttp.ClientSession() as session:
            # loop until we got all
            while True:
                # do request
                async with session.get(url, params=params, headers=self._headers, timeout=10) as response:
                    if response.status != 200:
                        raise ValueError("Could not query frames")

                    # create frames and return them
                    res = await response.json()

                    # create frames
                    new_frames = [PyobsArchiveFrameInfo(frame) for frame in res["results"]]
                    frames.extend(new_frames)

                    # got all?
                    if len(frames) >= res["count"]:
                        return cast(List[FrameInfo], frames)

                    # get next chunk
                    params["offset"] += len(new_frames)

    @staticmethod
    def _build_query(
        start: Optional[Time] = None,
        end: Optional[Time] = None,
        night: Optional[str] = None,
        site: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        image_type: Optional[ImageType] = None,
        binning: Optional[str] = None,
        filter_name: Optional[str] = None,
        rlevel: Optional[int] = None,
    ) -> Dict[str, Any]:
        # build params
        params: Dict[str, Any] = {}
        if start is not None:
            params["start"] = start.isot
        if end is not None:
            params["end"] = end.isot
        if night is not None:
            params["night"] = night
        if site is not None:
            params["SITE"] = site
        if telescope is not None:
            params["TELESCOPE"] = telescope
        if instrument is not None:
            params["INSTRUMENT"] = instrument
        if image_type is not None:
            params["IMAGETYPE"] = image_type.value
        if binning is not None:
            params["binning"] = binning
        if filter_name is not None:
            params["FILTER"] = filter_name
        if rlevel is not None:
            params["RLEVEL"] = rlevel
        return params

    async def download_frames(self, infos: List[FrameInfo]) -> List[Image]:
        # loop infos
        images = []
        for info in infos:
            # make sure it's the correct FrameInfo
            if not isinstance(info, PyobsArchiveFrameInfo):
                log.warning("Incorrect type for frame info.")
                continue

            # download
            url = urllib.parse.urljoin(self._url, info.url)
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=self._headers, timeout=60) as response:
                    if response.status != 200:
                        log.exception("Error downloading file %s.", info.filename)

                    # create image
                    try:
                        image = Image.from_bytes(await response.read())
                        images.append(image)
                    except OSError:
                        log.exception("Error downloading file %s.", info.filename)

        # return all
        return images

    async def download_headers(self, infos: List[PyobsArchiveFrameInfo]) -> List[Dict[str, Any]]:
        # loop infos
        headers = []
        for info in infos:
            # download
            url = urllib.parse.urljoin(self._url, info.url).replace("download", "headers")
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=self._headers, timeout=60) as response:
                    if response.status != 200:
                        log.error("Could not fetch headers for %s.", info.filename)

                    try:
                        results = (await response.json())["results"]
                        headers.append(dict((d["key"], d["value"]) for d in results))
                    except KeyError:
                        log.error("Could not fetch headers for %s.", info.filename)
                        headers.append({})

        # return all
        return headers

    async def upload_frames(self, images: List[Image]) -> None:
        # build URL
        url = urllib.parse.urljoin(self._url, "frames/create/")

        # do some initial GET request for getting the csrftoken
        async with aiohttp.ClientSession() as session:
            # do some initial GET request for getting the csrftoken
            data = aiohttp.FormData()
            async with session.get(self._url, headers=self._headers) as response:
                token = response.cookies["csrftoken"].value
                data.add_field("csrfmiddlewaretoken", token)

            for i, img in enumerate(images, 1):
                # get filename
                filename = img.header["FNAME"]

                # write HDU to BytesIO
                with io.BytesIO() as bio:
                    # write it
                    img.writeto(bio)
                    data.add_field(f"file{i}", bio.getvalue(), filename=filename)

            # post it
            async with session.post(url, data=data, timeout=10, headers=self._headers) as response:
                # success, if status code is 200
                if response.status != 200:
                    raise ValueError("Cannot write file, received status_code %d." % response.status)

                # check json
                json = await response.json()
                if "created" not in json or json["created"] == 0:
                    if "errors" in json:
                        raise ValueError("Could not create file in archive: " + str(json["errors"]))
                    else:
                        raise ValueError("Could not create file in archive.")


__all__ = ["PyobsArchiveFrameInfo", "PyobsArchive"]



import asyncio
import io
import json
import threading

import pandas as pd
import logging
import tornado.gen
import tornado.web
import tornado.ioloop

from .publisher import Publisher


log = logging.getLogger(__name__)


class LatestJsonHandler(tornado.web.RequestHandler):
    """Latest data as JSON."""

    @tornado.gen.coroutine
    def get(self):
        # set header
        self.set_header("content-type", "application/json")

        # no data?
        data = self.application.data
        if len(data) == 0:
            self.finish()
            return

        # convert last row to dict
        row = data.iloc[-1].to_dict()

        self.set_header("content-type", "application/json")
        self.write(json.dumps(row))
        self.finish()


class HistoryCsvHandler(tornado.web.RequestHandler):
    """History as CSV."""

    @tornado.gen.coroutine
    def get(self):
        # set header
        self.set_header("content-type", "text/csv")

        # write data
        with io.StringIO() as sio:
            self.application.data.to_csv(sio, index=False)
            self.write(sio.getvalue())
            self.finish()


class HttpPublisher(Publisher, tornado.web.Application):
    def __init__(self, port: int = 37077, keep: int = 10, **kwargs):
        """Initialize new CSV publisher.

        Args:
            filename: Name of file to log in.
            keep: Number of entries to keep.
        """
        Publisher.__init__(self, **kwargs)

        # add thread func
        self.add_background_task(self._http, False)

        # init tornado web server
        tornado.web.Application.__init__(
            self, [(r"/latest.json", LatestJsonHandler), (r"/history.csv", HistoryCsvHandler)]
        )

        # store stuff
        self._io_loop = None
        self._lock = threading.RLock()
        self._port = port
        self._keep = keep
        self._data = pd.DataFrame()

    def close(self):
        """Close server."""

        # close io loop and parent
        self._io_loop.add_callback(self._io_loop.stop)
        Publisher.close(self)

    @property
    def data(self):
        with self._lock:
            return self._data.copy()

    def _http(self):
        """Thread function for the web server."""

        # create io loop
        asyncio.set_event_loop(asyncio.new_event_loop())
        self._io_loop = tornado.ioloop.IOLoop.current()
        self._io_loop.make_current()

        # start listening
        log.info("Starting HTTP server on port %d.", self._port)
        self.listen(self._port)

        # start the io loop
        self._io_loop.start()

    def __call__(self, **kwargs):
        """Publish the given results.

        Args:
            **kwargs: Results to publish.
        """

        # lock
        with self._lock:
            # append data
            row = pd.DataFrame(kwargs, index=[0])
            self._data = pd.concat([self._data, row], ignore_index=True)

            # truncate
            if len(self.data) > self._keep:
                self._data = self._data.iloc[-self._keep :]


__all__ = ["HttpPublisher"]



import asyncio
import logging
import threading
from typing import Any, Union

from pyobs.utils.parallel import acquire_lock

log = logging.getLogger(__name__)


class AcquireLockFailed(Exception):
    pass


class LockWithAbort(object):
    """Tries to acquire a lock. If unsuccessful, it sets the event and tries again."""

    def __init__(self, lock: asyncio.Lock, event: asyncio.Event):
        self.lock = lock
        self.event = event
        self.acquired = False

    async def __aenter__(self) -> None:
        # first try to acquire lock without timeout
        self.acquired = await acquire_lock(self.lock)

        # not successful?
        if not self.acquired:
            # could not acquire lock, so set abort_event
            self.event.set()

            # try to acquire again with a timeout
            self.acquired = await acquire_lock(self.lock, 10.0)

            # still not successful?
            if not self.acquired:
                # raise exception
                raise AcquireLockFailed()

        # got lock, so unset abort and remember that we were successful
        self.event.clear()

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        # if we acquired the lock, we release it again here
        if self.acquired:
            self.lock.release()
            self.acquired = False


__all__ = ["LockWithAbort", "AcquireLockFailed"]



import asyncio
import logging
from typing import Tuple, Dict, Any, Optional
import astropy.units as u

from pyobs.images.meta import OnSkyDistance
from pyobs.images.meta.exptime import ExpTime
from pyobs.interfaces import IAcquisition
from pyobs.modules import Module
from pyobs.mixins import CameraSettingsMixin
from pyobs.modules import timeout
from pyobs.utils.enums import ImageType
from pyobs.utils.publisher import CsvPublisher
from pyobs.utils.time import Time
from ._base import BasePointing
from ...interfaces import IExposureTime, IImageType, ITelescope, IImageGrabber, IOffsetsRaDec, IOffsetsAltAz, ICamera

log = logging.getLogger(__name__)


class Acquisition(BasePointing, CameraSettingsMixin, IAcquisition):
    """Class for telescope acquisition."""

    __module__ = "pyobs.modules.pointing"

    def __init__(
        self,
        exposure_time: float,
        target_pixel: Optional[Tuple[float, float]] = None,
        attempts: int = 5,
        tolerance: float = 1,
        max_offset: float = 120,
        log_file: Optional[str] = None,
        **kwargs: Any,
    ):
        """Create a new acquisition.

        Args:
            exposure_time: Default exposure time.
            target_pixel: (x, y) tuple of pixel that the star should be positioned on. If None, center of image is used.
            attempts: Number of attempts before giving up.
            tolerance: Tolerance in position to reach in arcsec.
            max_offset: Maximum offset to move in arcsec.
            log_file: Name of file to write log to.
        """
        BasePointing.__init__(self, **kwargs)

        # store
        self._default_exposure_time = exposure_time
        self._is_running = False
        self._target_pixel = target_pixel
        self._attempts = attempts
        self._tolerance = tolerance * u.arcsec
        self._max_offset = max_offset * u.arcsec

        # init log file
        self._publisher = CsvPublisher(log_file) if log_file is not None else None

        # init camera settings mixin
        CameraSettingsMixin.__init__(self, **kwargs)

    async def open(self) -> None:
        """Open module"""
        await Module.open(self)

        # check telescope and camera
        try:
            await self.proxy(self._telescope, ITelescope)
            await self.proxy(self._camera, ICamera)
        except ValueError:
            log.warning("Either camera or telescope do not exist or are not of correct type at the moment.")

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether a service is running."""
        return self._is_running

    @timeout(120)
    async def acquire_target(self, **kwargs: Any) -> Dict[str, Any]:
        """Acquire target at given coordinates.

        If no RA/Dec are given, start from current position. Might not work for some implementations that require
        coordinates.

        Returns:
            A dictionary with entries for datetime, ra, dec, alt, az, and either off_ra, off_dec or off_alt, off_az.

        Raises:
            ValueError: If target could not be acquired.
        """

        try:
            self._is_running = True
            return await self._acquire(self._default_exposure_time)
        finally:
            self._is_running = False

    async def _acquire(self, exposure_time: float) -> Dict[str, Any]:
        """Actually acquire target."""

        # get telescope
        log.info("Getting proxy for telescope...")
        telescope = await self.proxy(self._telescope, ITelescope)

        # get camera
        log.info("Getting proxy for camera...")
        camera = await self.proxy(self._camera, IImageGrabber)

        # do camera settings
        await self._do_camera_settings(camera)

        # try given number of attempts
        for a in range(self._attempts):
            # set exposure time and image type and take image
            if isinstance(camera, IExposureTime):
                log.info("Exposing image for %.1f seconds...", exposure_time)
                await camera.set_exposure_time(exposure_time)
            else:
                log.info("Exposing image...")
            if isinstance(camera, IImageType):
                await camera.set_image_type(ImageType.ACQUISITION)
            filename = await camera.grab_image()

            # download image
            log.info("Downloading image...")
            if filename is None:
                log.warning("Did not receive an image.")
                continue
            image = await self.vfs.read_image(filename)

            # get offset
            log.info("Analysing image...")
            loop = asyncio.get_running_loop()
            image = await self.run_pipeline(image)

            # calculate distance from offset
            osd = image.get_meta(OnSkyDistance)
            if osd is None:
                log.warning("No on sky distance found in meta.")
                continue
            log.info("Found a distance to target of %.2f arcsec.", osd.distance.arcsec)

            # get distance
            if osd.distance < self._tolerance:
                # we're finished!
                log.info("Target successfully acquired.")

                # get current Alt/Az
                cur_alt, cur_az = await telescope.get_altaz()
                cur_ra, cur_dec = await telescope.get_radec()

                # prepare log entry
                log_entry = {"datetime": Time.now().isot, "ra": cur_ra, "dec": cur_dec, "alt": cur_alt, "az": cur_az}

                # Alt/Az or RA/Dec?
                if isinstance(telescope, IOffsetsRaDec):
                    log_entry["off_ra"], log_entry["off_dec"] = await telescope.get_offsets_radec()
                elif isinstance(telescope, IOffsetsAltAz):
                    log_entry["off_alt"], log_entry["off_az"] = await telescope.get_offsets_altaz()

                # write log
                # TODO: reactivate!
                # if self._publisher is not None:
                #    await self._publisher(**log_entry)

                # finished
                return log_entry

            # abort?
            if osd.distance > self._max_offset:
                # move a maximum of 120"=2'
                raise ValueError("Calculated offsets too large.")

            # apply offsets
            if await self._apply(image, telescope, self.location):
                log.info("Finished image.")
            else:
                log.warning("Could not apply offsets.")

            # new exposure time?
            if image.has_meta(ExpTime):
                exposure_time = image.get_meta(ExpTime).exptime

        # could not acquire target
        raise ValueError("Could not acquire target within given tolerance.")


__all__ = ["Acquisition"]



from abc import ABCMeta, abstractmethod

from pyobs.images import Image
from pyobs.images.processor import ImageProcessor


class SourceDetection(ImageProcessor, metaclass=ABCMeta):
    """Base class for source detection."""

    __module__ = "pyobs.images.processors.detection"

    @abstractmethod
    async def __call__(self, image: Image) -> Image:
        """Find stars in given image and append catalog.

        Args:
            image: Image to find stars in.

        Returns:
            Image with attached catalog.
        """
        ...


__all__ = ["SourceDetection"]



from abc import ABCMeta, abstractmethod
from typing import Optional, Any, List
from astroplan import ObservingBlock

from pyobs.utils.time import Time
from pyobs.object import Object


class TaskArchive(Object, metaclass=ABCMeta):
    def __init__(self, **kwargs: Any):
        Object.__init__(self, **kwargs)

    @abstractmethod
    async def last_changed(self) -> Optional[Time]:
        """Returns time when last time any blocks changed."""
        ...

    @abstractmethod
    async def get_schedulable_blocks(self) -> List[ObservingBlock]:
        """Returns list of schedulable blocks.

        Returns:
            List of schedulable blocks
        """
        ...


__all__ = ["TaskArchive"]



from abc import ABCMeta, abstractmethod
from typing import Any, Dict

from .interface import Interface


class ITemperatures(Interface, metaclass=ABCMeta):
    """The module can return temperatures measured on some device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_temperatures(self, **kwargs: Any) -> Dict[str, float]:
        """Returns all temperatures measured by this module.

        Returns:
            Dict containing temperatures.
        """
        ...


__all__ = ["ITemperatures"]



from slixmpp.plugins.xep_0009 import XEP_0009 as XEP_0009_original
from slixmpp.xmlstream import ET


class XEP_0009(XEP_0009_original):
    """Small fix for the original XEP_0009 plugin."""

    def _handle_error(self, iq):
        pass

    def extract_method(self, stanza):
        xml = ET.fromstring("%s" % stanza)
        return xml.find("./{jabber:iq:rpc}methodCall/{jabber:iq:rpc}methodName").text

    def item_not_found(self, iq):
        """Expose method to public."""
        return self._item_not_found(iq)

    def send_fault(self, iq, fault_xml):
        """Expose method to public."""
        return self._send_fault(iq, fault_xml)

    def make_iq_method_call(self, pto, pmethod, params):
        iq = self.xmpp.make_iq_set()
        iq["to"] = pto
        iq["from"] = self.xmpp.boundjid.full
        iq.enable("rpc_query")
        iq["rpc_query"]["method_call"]["method_name"] = pmethod
        iq["rpc_query"]["method_call"]["params"] = params
        return iq

    def make_iq_method_response(self, pid, pto, params):
        iq = self.xmpp.make_iq_result(pid)
        iq["to"] = pto
        iq["from"] = self.xmpp.boundjid.full
        iq.enable("rpc_query")
        iq["rpc_query"]["method_response"]["params"] = params
        return iq

    def make_iq_method_response_fault(self, pid, pto, params):
        iq = self.xmpp.make_iq_result(pid)
        iq["to"] = pto
        iq["from"] = self.xmpp.boundjid.full
        iq.enable("rpc_query")
        iq["rpc_query"]["method_response"]["params"] = None
        iq["rpc_query"]["method_response"]["fault"] = params
        return iq



from abc import ABCMeta, abstractmethod
from typing import Any

from .IImageGrabber import IImageGrabber


class IVideo(IImageGrabber, metaclass=ABCMeta):
    """The module controls a video streaming device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_video(self, **kwargs: Any) -> str:
        """Returns path to video.

        Returns:
            Path to video.
        """
        ...


__all__ = ["IVideo"]



# taken from: https://www.peterbe.com/plog/best-practice-with-retries-with-requests
from typing import Iterable, Optional

import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

# disable warnings
# TODO: is there a better way? we can link to actual CA:
# https://urllib3.readthedocs.io/en/latest/user-guide.html#ssl
# but does that only work in Linux?
# some more:
# https://github.com/psf/requests/issues/2214
import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def requests_retry_session(
    retries: int = 3,
    backoff_factor: float = 0.3,
    status_forcelist: Iterable[int] = (500, 502, 504),
    session: Optional[requests.Session] = None,
) -> requests.Session:
    session = session or requests.Session()
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
    )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount("http://", adapter)
    session.mount("https://", adapter)
    return session



import asyncio
import logging
from typing import Dict, Any, Union

from pyobs.interfaces import IRunnable
from pyobs.interfaces import IFlatField, IFilters, IBinning
from pyobs.modules import Module
from pyobs.modules import timeout
from pyobs.object import get_object
from pyobs.utils.parallel import event_wait
from pyobs.utils.skyflats.priorities.base import SkyflatPriorities
from pyobs.utils.skyflats.scheduler import Scheduler, SchedulerItem
from pyobs.utils.time import Time

log = logging.getLogger(__name__)


class FlatFieldScheduler(Module, IRunnable):
    """Run the flat-field scheduler."""

    __module__ = "pyobs.modules.flatfield"

    def __init__(
        self,
        flatfield: Union[str, IFlatField],
        functions: Dict[str, str],
        priorities: Union[Dict[str, Any], SkyflatPriorities],
        min_exptime: float = 0.5,
        max_exptime: float = 5,
        timespan: float = 7200,
        filter_change: float = 30,
        count: int = 20,
        **kwargs: Any,
    ):
        """Initialize a new flat field scheduler.

        Args:
            flatfield: Flat field module to use
            functions: Dict with flat functions
            priorities: Class handling priorities
            min_exptime: Minimum exposure time [s]
            max_exptime: Maximum exposure time [s]
            timespan: Time to scheduler after start [s]
            filter_change: Time required for filter change [s]
            count: Number of flats to take per filter/binning
        """
        Module.__init__(self, **kwargs)

        # store
        self._flatfield = flatfield
        self._count = count
        self._running = False

        # abort
        self._abort = asyncio.Event()

        # priorities
        prio = get_object(priorities, SkyflatPriorities)

        # create scheduler
        self._scheduler = Scheduler(
            functions,
            prio,
            self.observer,
            min_exptime=min_exptime,
            max_exptime=max_exptime,
            timespan=timespan,
            filter_change=filter_change,
            count=count,
        )

    async def open(self) -> None:
        """Open module"""
        await Module.open(self)

        # check flat field
        try:
            await self.proxy(self._flatfield, IFlatField)
        except ValueError:
            log.warning("Flatfield module does not exist or is not of correct type at the moment.")

    @timeout(7200)
    async def run(self, **kwargs: Any) -> None:
        """Perform flat-fielding"""

        # check
        if self._running:
            raise ValueError("Already running.")
        self._running = True

        try:
            # start
            log.info("Performing flat fielding...")
            self._abort = asyncio.Event()

            # get flat fielder
            log.info("Getting proxy for flat fielder...")
            flatfield = await self.proxy(self._flatfield, IFlatField)

            # do schedule
            log.info("Scheduling flats...")
            await self._scheduler(Time.now())

            # do flat fields
            sched: SchedulerItem
            for item in self._scheduler:
                # aborted?
                if self._abort.is_set():
                    log.info("Scheduler aborted.")
                    break

                # start
                log.info("Taking %d flats in %s %dx%d...", self._count, item.filter_name, item.binning, item.binning)
                if isinstance(flatfield, IFilters):
                    await flatfield.set_filter(item.filter_name)
                if isinstance(flatfield, IBinning):
                    await flatfield.set_binning(*item.binning)
                future = asyncio.create_task(flatfield.flat_field(self._count))

                # wait for it
                while not future.done():
                    # aborted?
                    if self._abort.is_set():
                        log.info("Aborting current flat field...")
                        await flatfield.abort()

                    # sleep a little
                    await event_wait(self._abort, 1)

            # finished
            log.info("Finished.")

        finally:
            self._running = False

    @timeout(20)
    async def abort(self, **kwargs: Any) -> None:
        """Abort current actions."""
        self._abort.set()


__all__ = ["FlatFieldScheduler"]



import logging
from typing import Any, Optional, List

from pyobs.modules import Module
from pyobs.events import MotionStatusChangedEvent
from pyobs.utils.enums import MotionStatus

log = logging.getLogger(__name__)


class MotionStatusMixin:
    """Mixin for IMotion devices for handling status."""

    __module__ = "pyobs.mixins"

    def __init__(self, motion_status_interfaces: Optional[List[str]] = None, **kwargs: Any):
        """Initializes the mixin.

        Args:
            interfaces: List of interfaces to handle or None
        """
        self.__motion_status_interfaces = [] if motion_status_interfaces is None else motion_status_interfaces
        self.__motion_status = MotionStatus.UNKNOWN
        self.__motion_status_single = {i: MotionStatus.UNKNOWN for i in self.__motion_status_interfaces}

    async def open(self) -> None:
        # subscribe to events
        if isinstance(self, Module) and self.comm:
            await self.comm.register_event(MotionStatusChangedEvent)

    async def _change_motion_status(self, status: MotionStatus, interface: Optional[str] = None) -> None:
        """Change motion status and send event,

        Args:
            status: New motion status
            interface: Interface to set motion status for
        """

        # did something change?
        changed = False

        # global or individual?
        if interface is None:
            # did status change?
            if self.__motion_status != status:
                # set it
                changed = True
                self.__motion_status = status

                # also set all individual interfaces
                for i in self.__motion_status_interfaces:
                    if self.__motion_status_single[i] != status:
                        changed = True
                        self.__motion_status_single[i] = status

        else:
            # does it exist?
            if interface not in self.__motion_status_interfaces:
                return

            # did status change?
            if self.__motion_status_single[interface] != status:
                # set it
                self.__motion_status_single[interface] = status
                changed = True

                # combine status
                self.__motion_status = self._combine_motion_status()

        # send event
        if changed:
            this = self

            # log it
            if interface is None:
                log.info("Changed motion status to %s.", status)
            else:
                log.info("Changed motion status of %s to %s.", interface, status)

            # send event
            if not isinstance(self, Module):
                raise ValueError("This is not a module.")
            await self.comm.send_event(
                MotionStatusChangedEvent(status=this.__motion_status, interfaces=this.__motion_status_single)
            )

    def _combine_motion_status(self) -> MotionStatus:
        """Method for combining motion statuses for individual interfaces into the global one. Can be overriden."""

        # none?
        if len(self.__motion_status_interfaces) == 0:
            return MotionStatus.UNKNOWN

        # if any interface is of state ERROR, UNKNOWN, INITIALIZING, PARKING, SLEWING
        # we use that as global status (in that order)
        for status in [
            MotionStatus.ERROR,
            MotionStatus.UNKNOWN,
            MotionStatus.INITIALIZING,
            MotionStatus.PARKING,
            MotionStatus.SLEWING,
        ]:
            if status in self.__motion_status_single.values():
                return status

        # otherwise just take status of first interface
        return self.__motion_status_single[self.__motion_status_interfaces[0]]

    async def get_motion_status(self, device: Optional[str] = None, **kwargs: Any) -> MotionStatus:
        """Returns current motion status.

        Args:
            device: Name of device to get status for, or None.

        Returns:
            A string from the Status enumerator.
        """

        # global or individual?
        if device is None:
            return self.__motion_status

        else:
            # does it exist?
            if device in self.__motion_status_single:
                return self.__motion_status_single[device]
            else:
                raise KeyError


__all__ = ["MotionStatusMixin"]



import asyncio
import logging
from typing import Any, Optional

from pyobs.events import RoofOpenedEvent, RoofClosingEvent
from pyobs.interfaces import IRoof
from pyobs.modules import timeout
from pyobs.modules.roof import BaseRoof
from pyobs.utils.enums import MotionStatus
from pyobs.utils.threads import LockWithAbort

log = logging.getLogger(__name__)


class DummyRoof(BaseRoof, IRoof):
    """A dummy camera for testing."""

    __module__ = "pyobs.modules.roof"

    def __init__(self, **kwargs: Any):
        """Creates a new dummy root."""
        BaseRoof.__init__(self, **kwargs)

        # dummy state
        self.open_percentage = 0

        # allow to abort motion
        self._lock_motion = asyncio.Lock()
        self._abort_motion = asyncio.Event()

    async def open(self) -> None:
        """Open module."""
        await BaseRoof.open(self)

        # register event
        await self.comm.register_event(RoofOpenedEvent)
        await self.comm.register_event(RoofClosingEvent)

    @timeout(15)
    async def init(self, **kwargs: Any) -> None:
        """Open the roof.

        Raises:
            AcquireLockFailed: If current motion could not be aborted.
        """

        # already open?
        if self.open_percentage != 100:
            # acquire lock
            with LockWithAbort(self._lock_motion, self._abort_motion):
                # change status
                await self._change_motion_status(MotionStatus.INITIALIZING)

                # open roof
                while self.open_percentage < 100:
                    # open more
                    self.open_percentage += 1

                    # abort?
                    if self._abort_motion.is_set():
                        await self._change_motion_status(MotionStatus.IDLE)
                        return

                    # wait a little
                    await asyncio.sleep(0.1)

                # open fully
                self.open_percentage = 100

                # change status
                await self._change_motion_status(MotionStatus.IDLE)

                # send event
                self.comm.send_event(RoofOpenedEvent())

    @timeout(15)
    async def park(self, **kwargs: Any) -> None:
        """Close the roof.

        Raises:
            AcquireLockFailed: If current motion could not be aborted.
        """

        # already closed?
        if self.open_percentage != 0:
            # acquire lock
            with LockWithAbort(self._lock_motion, self._abort_motion):
                # change status
                await self._change_motion_status(MotionStatus.PARKING)

                # send event
                self.comm.send_event(RoofClosingEvent())

                # close roof
                while self.open_percentage > 0:
                    # close more
                    self.open_percentage -= 1

                    # abort?
                    if self._abort_motion.is_set():
                        await self._change_motion_status(MotionStatus.IDLE)
                        return

                    # wait a little
                    await asyncio.sleep(0.1)

                # change status
                await self._change_motion_status(MotionStatus.PARKED)

    def get_percent_open(self) -> float:
        """Get the percentage the roof is open."""
        return self.open_percentage

    async def stop_motion(self, device: Optional[str] = None, **kwargs: Any) -> None:
        """Stop the motion.

        Args:
            device: Name of device to stop, or None for all.

        Raises:
            AcquireLockFailed: If current motion could not be aborted.
        """

        # change status
        await self._change_motion_status(MotionStatus.ABORTING)

        # abort
        # acquire lock
        with LockWithAbort(self._lock_motion, self._abort_motion):
            # change status
            await self._change_motion_status(MotionStatus.IDLE)


__all__ = ["DummyRoof"]



import logging
from abc import ABCMeta
from typing import List, Dict, Tuple, Any, Optional

from pyobs.interfaces import IDome
from pyobs.utils import exceptions as exc
from .baseroof import BaseRoof


log = logging.getLogger(__name__)


class BaseDome(IDome, BaseRoof, metaclass=ABCMeta):
    """Base class for domes."""

    __module__ = "pyobs.modules.roof"

    def __init__(self, **kwargs: Any):
        """Initialize a new base dome."""
        BaseRoof.__init__(self, **kwargs)

        # register exception
        exc.register_exception(exc.MotionError, 3, timespan=600, callback=self._default_remote_error_callback)

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # get from parent
        hdr = await BaseRoof.get_fits_header_before(self, namespaces, **kwargs)

        # add azimuth and return it
        _, az = await self.get_altaz()
        hdr["ROOF-AZ"] = (az, "Azimuth of roof slit, deg E of N")
        return hdr


__all__ = ["BaseDome"]



from __future__ import annotations
import json
import time
import uuid
import logging
from typing import Dict, Any, Optional


log = logging.getLogger(__name__)


class Event:
    """Base class for all events."""

    __module__ = "pyobs.events"
    local = False

    def __init__(self, **kwargs: Any):
        self.uuid = str(uuid.uuid4())
        self.timestamp = time.time()
        self.data: Any = {}

    def to_json(self) -> Dict[str, Any]:
        """JSON representation of event."""
        return {"type": self.__class__.__name__, "timestamp": self.timestamp, "uuid": self.uuid, "data": self.data}

    def __str__(self) -> str:
        """String representation of event."""
        return json.dumps(self.to_json())

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        """Generic from_dict method for derived classes that don't need their own."""
        return cls(**d)


class EventFactory(object):
    packages = [__package__]

    @staticmethod
    def from_dict(obj_dict: Dict[str, Any]) -> Optional[Event]:
        """Create Event from a dictionary.

        Args:
            obj_dict: JSON string for event.

        Returns:
            Event object containing event.
        """

        # create class
        cls: Optional[Event] = None
        for p in EventFactory.packages:
            # import package
            parts = p.split(".")
            pkg = __import__(parts[0])
            for comp in parts[1:]:
                pkg = getattr(pkg, comp)

            # does it have the given type as class?
            if hasattr(pkg, obj_dict["type"]):
                cls = getattr(pkg, obj_dict["type"])
                break

        # not found?
        if cls is None:
            return None

        # instantiate object and set data
        try:
            kwargs = obj_dict["data"] if "data" in obj_dict and obj_dict["data"] is not None else {}
            if not isinstance(kwargs, dict):
                raise ValueError('Invalid event structure for event %s: "%s".' % (str(cls), str(kwargs)))
            obj: Event = cls.from_dict(kwargs)
            obj.uuid = obj_dict["uuid"]
            obj.timestamp = obj_dict["timestamp"] if "timestamp" in obj_dict else 0
            return obj

        except ValueError:
            log.warning("Could not create event %s." % obj_dict["type"])
            return None


__all__ = ["Event", "EventFactory"]



import asyncio
import datetime
import logging
import warnings
from abc import ABCMeta, abstractmethod
from typing import Tuple, Optional, Dict, Any, NamedTuple, List
import numpy as np
from astropy.io import fits
from numpy.typing import NDArray

from pyobs.mixins.fitsheader import ImageFitsHeaderMixin
from pyobs.utils.enums import ImageType, ExposureStatus
from pyobs.images import Image
from pyobs.modules import Module
from pyobs.events import NewImageEvent, ExposureStatusChangedEvent
from pyobs.interfaces import ICamera, IExposureTime, IImageType
from pyobs.modules import timeout
from pyobs.utils import exceptions as exc

log = logging.getLogger(__name__)


class CameraException(Exception):
    pass


class ExposureInfo(NamedTuple):
    """Info about a running exposure."""

    start: datetime.datetime
    exposure_time: float


async def calc_expose_timeout(camera: IExposureTime, *args: Any, **kwargs: Any) -> float:
    """Calculates timeout for expose()."""
    return await camera.get_exposure_time() + 30


class BaseCamera(Module, ImageFitsHeaderMixin, ICamera, IExposureTime, IImageType, metaclass=ABCMeta):
    """Base class for all camera modules."""

    __module__ = "pyobs.modules.camera"

    def __init__(
        self,
        fits_headers: Optional[Dict[str, Any]] = None,
        centre: Optional[Tuple[float, float]] = None,
        rotation: float = 0.0,
        flip: bool = False,
        filenames: str = "/cache/pyobs-{DAY-OBS|date:}-{FRAMENUM|string:04d}-{IMAGETYP|type}00.fits.gz",
        fits_namespaces: Optional[List[str]] = None,
        **kwargs: Any,
    ):
        """Creates a new BaseCamera.

        Args:
            fits_headers: Additional FITS headers.
            centre: (x, y) tuple of camera centre.
            rotation: Rotation east of north.
            flip: Whether or not to flip the image along its first axis.
            filenames: Template for file naming.
            fits_namespaces: List of namespaces for FITS headers that this camera should request
        """
        Module.__init__(self, **kwargs)
        ImageFitsHeaderMixin.__init__(
            self,
            fits_namespaces=fits_namespaces,
            fits_headers=fits_headers,
            centre=centre,
            rotation=rotation,
            filenames=filenames,
        )

        # check
        if self.comm is None:
            log.warning("No comm module given, will not be able to signal new images!")

        # store
        self._flip = flip
        self._exposure_time: float = 0.0
        self._image_type = ImageType.OBJECT

        # init camera
        self._exposure: Optional[ExposureInfo] = None
        self._camera_status = ExposureStatus.IDLE

        # multi-threading
        self.expose_abort = asyncio.Event()

        # register exception
        exc.register_exception(exc.GrabImageError, 3, timespan=600, callback=self._default_remote_error_callback)

    async def open(self) -> None:
        """Open module."""
        await Module.open(self)

        # subscribe to events
        if self.comm:
            await self.comm.register_event(NewImageEvent)
            await self.comm.register_event(ExposureStatusChangedEvent)

    async def set_exposure_time(self, exposure_time: float, **kwargs: Any) -> None:
        """Set the exposure time in seconds.

        Args:
            exposure_time: Exposure time in seconds.

        Raises:
            ValueError: If exposure time could not be set.
        """
        log.info("Setting exposure time to %.5fs...", exposure_time)
        self._exposure_time = exposure_time

    async def get_exposure_time(self, **kwargs: Any) -> float:
        """Returns the exposure time in seconds.

        Returns:
            Exposure time in seconds.
        """
        return self._exposure_time

    async def set_image_type(self, image_type: ImageType, **kwargs: Any) -> None:
        """Set the image type.

        Args:
            image_type: New image type.
        """
        log.info("Setting image type to %s...", image_type)
        self._image_type = image_type

    async def get_image_type(self, **kwargs: Any) -> ImageType:
        """Returns the current image type.

        Returns:
            Current image type.
        """
        return self._image_type

    async def _change_exposure_status(self, status: ExposureStatus) -> None:
        """Change exposure status and send event,

        Args:
            status: New exposure status.
        """

        # send event, if it changed
        if self._camera_status != status:
            await self.comm.send_event(ExposureStatusChangedEvent(last=self._camera_status, current=status))

        # set it
        self._camera_status = status

    async def get_exposure_status(self, **kwargs: Any) -> ExposureStatus:
        """Returns the current status of the camera, which is one of 'idle', 'exposing', or 'readout'.

        Returns:
            Current status of camera.
        """
        return self._camera_status

    async def get_exposure_time_left(self, **kwargs: Any) -> float:
        """Returns the remaining exposure time on the current exposure in seconds.

        Returns:
            Remaining exposure time in seconds.
        """

        # if we're not exposing, there is nothing left
        if self._exposure is None:
            return 0.0

        # calculate difference between start of exposure and now, and return in ms
        duration = datetime.timedelta(seconds=self._exposure.exposure_time)
        diff = self._exposure.start + duration - datetime.datetime.utcnow()
        return diff.total_seconds()

    async def get_exposure_progress(self, **kwargs: Any) -> float:
        """Returns the progress of the current exposure in percent.

        Returns:
            Progress of the current exposure in percent.
        """

        # if we're not exposing, there is no progress
        if self._exposure is None:
            return 0.0

        # calculate difference between start of exposure and now
        diff = datetime.datetime.utcnow() - self._exposure[0]

        # zero exposure time?
        if self._exposure.exposure_time == 0.0 or self._camera_status == ExposureStatus.READOUT:
            return 100.0
        else:
            # return max of 100
            percentage = diff.total_seconds() / self._exposure[1] * 100.0
            return min(percentage, 100.0)

    @abstractmethod
    async def _expose(self, exposure_time: float, open_shutter: bool, abort_event: asyncio.Event) -> Image:
        """Actually do the exposure, should be implemented by derived classes.

        Args:
            exposure_time: The requested exposure time in seconds.
            open_shutter: Whether or not to open the shutter.
            abort_event: Event that gets triggered when exposure should be aborted.

        Returns:
            The actual image.

        Raises:
            GrabImageError: If exposure was not successful.
        """
        ...

    async def __expose(self, exposure_time: float, image_type: ImageType, broadcast: bool) -> Tuple[Image, str]:
        """Wrapper for a single exposure.

        Args:
            exposure_time: The requested exposure time in seconds.
            image_type: Type of image.
            broadcast: Whether or not the new image should be broadcasted.

        Returns:
            Tuple of the image itself and its filename.

        Raises:
            GrabImageError: If there was a problem grabbing the image.
        """

        # request fits headers
        header_futures_before = await self.request_fits_headers(before=True)

        # open the shutter?
        open_shutter = image_type not in [ImageType.BIAS, ImageType.DARK]

        # reset abort event
        self.expose_abort.clear()

        # do the exposure
        self._exposure = ExposureInfo(start=datetime.datetime.utcnow(), exposure_time=exposure_time)
        try:
            image = await self._expose(exposure_time, open_shutter, abort_event=self.expose_abort)
            if image is None or image.data is None:
                raise exc.GrabImageError("Could not take image.")

        except exc.PyObsError:
            # exposure was not successful (aborted?), so reset everything and re-raise
            self._exposure = None
            raise

        except Exception as e:
            # exposure was not successful (aborted?), so reset everything and wrap exception
            self._exposure = None
            raise exc.GrabImageError(str(e))

        # request fits headers again
        header_futures_after = await self.request_fits_headers(before=False)

        # flip it?
        if self._flip:
            # do we have three dimensions in array? need this for deciding which axis to flip
            is_3d = len(image.data.shape) == 3

            # flip image and make contiguous again
            flipped: NDArray[Any] = np.flip(image.data, axis=1 if is_3d else 0)
            image.data = np.ascontiguousarray(flipped)

        # add HDU name
        image.header["EXTNAME"] = "SCI"

        # add image type
        image.header["IMAGETYP"] = image_type.value

        # add fits headers and format filename
        await self.add_requested_fits_headers(image, header_futures_before)
        await self.add_requested_fits_headers(image, header_futures_after)
        await self.add_fits_headers(image)
        filename = self.format_filename(image)

        # don't want to save?
        if filename is None:
            self._exposure = None
            raise exc.GrabImageError("No filename given.")

        # upload file
        try:
            log.info("Uploading image to file server...")
            await self.vfs.write_image(filename, image)
        except FileNotFoundError:
            raise ValueError("Could not upload image.")

        # broadcast image path
        if broadcast and self.comm:
            log.info("Broadcasting image ID...")
            await self.comm.send_event(NewImageEvent(filename, image_type))

        # return image and unique
        self._exposure = None
        log.info("Finished image %s.", filename)
        return image, filename

    @timeout(calc_expose_timeout)
    async def grab_image(self, broadcast: bool = True, **kwargs: Any) -> str:
        """Grabs an image ans returns reference.

        Args:
            broadcast: Broadcast existence of image.

        Returns:
            Name of image that was taken.

        Raises:
            GrabImageError: If there was a problem grabbing the image.
        """

        # are we exposing?
        if self._camera_status != ExposureStatus.IDLE:
            raise CameraException("Cannot start new exposure because camera is not idle.")
        await self._change_exposure_status(ExposureStatus.EXPOSING)

        # expose
        try:
            image, filename = await self.__expose(self._exposure_time, self._image_type, broadcast)
        finally:
            await self._change_exposure_status(ExposureStatus.IDLE)

        # check
        if image is None:
            raise exc.GrabImageError("Could not take image.")
        else:
            if filename is None:
                raise ValueError("Image has not been saved, so cannot be retrieved by filename.")

        # return filename
        return filename

    async def _abort_exposure(self) -> None:
        """Abort the running exposure. Should be implemented by derived class.

        Raises:
            ValueError: If an error occured.
        """
        pass

    async def abort(self, **kwargs: Any) -> None:
        """Aborts the current exposure and sequence.

        Returns:
            Success or not.
        """

        # set abort event
        log.info("Aborting current image and sequence...")
        self.expose_abort.set()

        # do camera-specific abort
        await self._abort_exposure()

        # wait until state is not EXPOSING anymore
        while await self.get_exposure_status() == ExposureStatus.EXPOSING:
            await asyncio.sleep(0.1)

    @staticmethod
    def set_biassec_trimsec(hdr: fits.Header, left: int, top: int, width: int, height: int) -> None:
        """Calculates and sets the BIASSEC and TRIMSEC areas.

        Args:
            hdr:    FITS header (in/out)
            left:   left edge of data area
            top:    top edge of data area
            width:  width of data area
            height: height of data area
        """

        # get image area in unbinned coordinates
        img_left = hdr["XORGSUBF"]
        img_top = hdr["YORGSUBF"]
        img_width = hdr["NAXIS1"] * hdr["XBINNING"]
        img_height = hdr["NAXIS2"] * hdr["YBINNING"]

        # get intersection
        is_left = max(left, img_left)
        is_right = min(left + width, img_left + img_width)
        is_top = max(top, img_top)
        is_bottom = min(top + height, img_top + img_height)

        # for simplicity we allow prescan/overscan only in one dimension
        if (left < is_left or left + width > is_right) and (top < is_top or top + height > is_bottom):
            log.warning("BIASSEC/TRIMSEC can only be calculated with a prescan/overscan on one axis only.")
            return

        # comments
        c1 = "Bias overscan area [x1:x2,y1:y2] (binned)"
        c2 = "Image area [x1:x2,y1:y2] (binned)"

        # rectangle empty?
        if is_right <= is_left or is_bottom <= is_top:
            # easy case, all is BIASSEC, no TRIMSEC at all
            hdr["BIASSEC"] = ("[1:%d,1:%d]" % (hdr["NAXIS1"], hdr["NAXIS2"]), c1)
            return

        # we got a TRIMSEC, calculate its binned and windowd coordinates
        is_left_binned = np.floor((is_left - hdr["XORGSUBF"]) / hdr["XBINNING"]) + 1
        is_right_binned = np.ceil((is_right - hdr["XORGSUBF"]) / hdr["XBINNING"])
        is_top_binned = np.floor((is_top - hdr["YORGSUBF"]) / hdr["YBINNING"]) + 1
        is_bottom_binned = np.ceil((is_bottom - hdr["YORGSUBF"]) / hdr["YBINNING"])

        # set it
        hdr["TRIMSEC"] = ("[%d:%d,%d:%d]" % (is_left_binned, is_right_binned, is_top_binned, is_bottom_binned), c2)
        hdr["DATASEC"] = ("[%d:%d,%d:%d]" % (is_left_binned, is_right_binned, is_top_binned, is_bottom_binned), c2)

        # now get BIASSEC -- whatever we do, we only take the last (!) one
        # which axis?
        if img_left + img_width > left + width:
            left_binned = np.floor((is_right - hdr["XORGSUBF"]) / hdr["XBINNING"]) + 1
            hdr["BIASSEC"] = ("[%d:%d,1:%d]" % (left_binned, hdr["NAXIS1"], hdr["NAXIS2"]), c1)
        elif img_left < left:
            right_binned = np.ceil((is_left - hdr["XORGSUBF"]) / hdr["XBINNING"])
            hdr["BIASSEC"] = ("[1:%d,1:%d]" % (right_binned, hdr["NAXIS2"]), c1)
        elif img_top + img_height > top + height:
            top_binned = np.floor((is_bottom - hdr["YORGSUBF"]) / hdr["YBINNING"]) + 1
            hdr["BIASSEC"] = ("[1:%d,%d:%d]" % (hdr["NAXIS1"], top_binned, hdr["NAXIS2"]), c1)
        elif img_top < top:
            bottom_binned = np.ceil((is_top - hdr["YORGSUBF"]) / hdr["YBINNING"])
            hdr["BIASSEC"] = ("[1:%d,1:%d]" % (hdr["NAXIS1"], bottom_binned), c1)

    async def list_binnings(self, **kwargs: Any) -> List[Tuple[int, int]]:
        """List available binnings.

        Returns:
            List of available binnings as (x, y) tuples.
        """

        warnings.warn(
            "The default implementation for list_binnings() in BaseCamera will be removed in future versions",
            DeprecationWarning,
        )
        return [(1, 1), (2, 2), (3, 3)]


__all__ = ["BaseCamera", "CameraException"]



from typing import Any, Dict, List, cast
from urllib.parse import urljoin

import aiohttp

from pyobs.utils.time import Time


class Portal:
    def __init__(self, url: str, token: str):
        self.url = url
        self.token = token

    async def _get(self, path: str, timeout: int = 10) -> Any:
        """Do a GET request on the portal.

        Args:
            url: URL to request.

        Returns:
            Response for request.

        Raises:
            RuntimeError if the call failed.
            TimeoutError if the call timed out.
        """

        # build header
        headers = {"Authorization": "Token " + self.token}

        async with aiohttp.ClientSession() as session:
            async with session.get(urljoin(self.url, path), headers=headers, timeout=timeout) as response:
                if response.status != 200:
                    raise RuntimeError("Invalid response from portal: " + await response.text())
                return await response.json()

    async def last_changed(self) -> Time:
        t = await self._get("/api/last_changed/")
        return Time(t["last_change_time"])

    async def last_scheduled(self) -> Time:
        t = await self._get("/api/last_scheduled/")
        return Time(t["last_schedule_time"])

    async def schedulable_requests(self) -> List[Dict[str, Any]]:
        req = await self._get("/api/requestgroups/schedulable_requests/")
        return cast(List[Dict[str, Any]], req)

    async def proposals(self) -> List[Dict[str, Any]]:
        req = await self._get("/api/proposals/")
        return cast(List[Dict[str, Any]], req["results"])

    async def instruments(self) -> Dict[str, Any]:
        req = await self._get("/api/instruments/")
        return cast(Dict[str, Any], req)



"""
Source extraction
-----------------
"""

from .photometry import Photometry
from .photutil import PhotUtilsPhotometry
from .pysep import SepPhotometry


__all__ = ["Photometry", "PhotUtilsPhotometry", "SepPhotometry"]



from typing import Optional, Any, Dict
from typing_extensions import TypedDict

from pyobs.utils.time import Time
from pyobs.events.event import Event


DataType = TypedDict("DataType", {"eta": Optional[str]})


class GoodWeatherEvent(Event):
    """Event to be sent on good weather."""

    __module__ = "pyobs.events"

    def __init__(self, eta: Optional[Time] = None, **kwargs: Any):
        """Initializes a new good weather event.

        Args:
            eta: Predicted ETA for when the telescope will be fully operational
        """
        Event.__init__(self)
        self.data: DataType = {"eta": eta.isot if eta is not None else None}

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> Event:
        # get eta
        eta: Optional[Time] = None
        if "eta" in d and isinstance(d["eta"], str):
            eta = Time(d["eta"])

        # return object
        return GoodWeatherEvent(eta=eta)

    @property
    def eta(self) -> Optional[Time]:
        return Time(self.data["eta"]) if self.data["eta"] is not None else None


__all__ = ["GoodWeatherEvent"]



from typing import Tuple, Optional, Any
import numpy as np
from astropy.coordinates import EarthLocation, SkyCoord
from astropy.wcs import WCS
import logging

from pyobs.images import Image
from pyobs.images.meta import PixelOffsets
from pyobs.interfaces import ITelescope, IPointingRaDec, IPointingAltAz
from pyobs.object import Object
from pyobs.utils.publisher import CsvPublisher
from pyobs.utils.time import Time


log = logging.getLogger(__name__)


class ApplyOffsets(Object):
    """Apply offsets from a given image to a given telescope."""

    __module__ = "pyobs.utils.offsets"

    def __init__(self, log_file: Optional[str] = None, log_absolute: bool = False, **kwargs: Any):
        Object.__init__(self, **kwargs)

        # init log file
        self._publisher = None if log_file is None else CsvPublisher(log_file)
        self._log_absolute = log_absolute

    async def __call__(self, image: Image, telescope: ITelescope, location: EarthLocation) -> bool:
        """Take the pixel offsets stored in the meta data of the image and apply them to the given telescope.

        Args:
            image: Image to take offsets and WCS from.
            telescope: Telescope to apply offsets to.
            location: Observer location on Earth.

        Returns:
            Whether offsets have been applied successfully.
        """
        return False

    def _get_radec_center_target(self, image: Image, location: EarthLocation) -> Tuple[SkyCoord, SkyCoord]:
        """Return RA/Dec of central pixel and of central pixel plus offsets.

        Args:
            image: Image to take header and offsets from.
            location: Observer location.

        Returns:
            Tuple of RA/Dec coordinates of center and of centre+offsets.
        """

        # get offsets
        offsets = image.get_meta(PixelOffsets)
        log.info("Found pixel shift of dx=%.2f, dy=%.2f.", offsets.dx, offsets.dy)

        # get reference pixel and date obs
        cx, cy = image.header["DET-CPX1"], image.header["DET-CPX2"]

        # get WCS and RA/DEC for pixel and pixel + dx/dy
        w = WCS(image.header)
        center = w.pixel_to_world(cx, cy)
        target = w.pixel_to_world(cx + offsets.dx, cy + offsets.dy)
        return center, target

    async def _log_offset(
        self,
        telescope: ITelescope,
        x_header: str,
        x_cur: float,
        x_delta: float,
        y_header: str,
        y_cur: float,
        y_delta: float,
    ) -> None:
        """Logs offset.

        Args:
            telescope: Telescope to use
            x_header: Header name for x value
            x_cur: Current x value
            x_delta: Delta for x
            y_header: Header name for y value
            y_cur: Current y value
            y_delta: Delta for y
        """

        # nothing?
        if self._publisher is None:
            return

        # init
        log_entry = {"datetime": Time.now().isot}

        # RA/Dec?
        if isinstance(telescope, IPointingRaDec):
            log_entry["ra"], log_entry["dec"] = await telescope.get_radec()

        # Alt/Az?
        if isinstance(telescope, IPointingAltAz):
            log_entry["alt"], log_entry["az"] = await telescope.get_altaz()

        # add entry
        log_entry[x_header] = x_delta + (x_cur if self._log_absolute else 0.0)
        log_entry[y_header] = y_delta + (y_cur if self._log_absolute else 0.0)

        # add separation (assume Euclidian space)
        log_entry["separation"] = np.sqrt(log_entry[x_header] ** 2 + log_entry[y_header] ** 2)

        # log it
        await self._publisher(**log_entry)


__all__ = ["ApplyOffsets"]



from __future__ import annotations
import logging
from typing import Dict, Any, Optional, Union, TYPE_CHECKING

from pyobs.interfaces import IFilters, IBinning, IFlatField, ITelescope, IRoof
from pyobs.robotic.scripts import Script
from pyobs.utils.skyflats.priorities.base import SkyflatPriorities
from pyobs.utils.skyflats.scheduler import Scheduler, SchedulerItem
from pyobs.utils.time import Time

if TYPE_CHECKING:
    from pyobs.robotic import TaskSchedule, TaskArchive, TaskRunner

log = logging.getLogger(__name__)


class SkyFlats(Script):
    """Script for scheduling and running skyflats using an IFlatField module."""

    def __init__(
        self,
        roof: Union[str, IRoof],
        telescope: Union[str, ITelescope],
        flatfield: Union[str, IFlatField],
        functions: Dict[str, Any],
        priorities: Union[Dict[str, Any], SkyflatPriorities],
        min_exptime: float = 0.5,
        max_exptime: float = 5,
        timespan: float = 7200,
        filter_change: float = 30,
        count: int = 20,
        readout: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ):
        """Init a new SkyFlats script.

        Args:
            roof: Roof to use
            telescope: Telescope to use
            flatfield: FlatFielder to use
            functions: Dict with solalt-exptime functions for all filters/binning
            priorities: SkyflatPriorities object that returns priorities
            min_exptime: Minimum exposure time for flats
            max_exptime: Maximum exposure time for flats
            timespan: Timespan from now that should be scheduled [s]
            filter_change: Time required for filter change [s]
            count: Number of flats to schedule
            readout: Dictionary with readout times (in sec) per binning (as BxB).
        """
        Script.__init__(self, **kwargs)

        # store modules
        self._roof = roof
        self._telescope = telescope
        self._flatfield = flatfield

        # stuff
        self._count = count

        # get archive and priorities
        prio = self.get_object(priorities, SkyflatPriorities)

        # create scheduler
        self._scheduler = Scheduler(
            functions,
            prio,
            self.observer,
            min_exptime=min_exptime,
            max_exptime=max_exptime,
            timespan=timespan,
            filter_change=filter_change,
            count=count,
            readout=readout,
        )

    async def can_run(self) -> bool:
        """Whether this config can currently run.

        Returns:
            True if script can run now.
        """

        # get modules
        try:
            roof = await self.comm.proxy(self._roof, IRoof)
            telescope = await self.comm.proxy(self._telescope, ITelescope)
            await self.comm.proxy(self._flatfield, IFlatField)
        except ValueError:
            return False

        # we need an open roof and a working telescope
        if not await roof.is_ready() or not await telescope.is_ready():
            return False

        # seems alright
        return True

    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
    ) -> None:
        """Run script.

        Raises:
            InterruptedError: If interrupted
        """

        # get proxy for flatfield
        flatfield = await self.comm.proxy(self._flatfield, IFlatField)

        # schedule
        log.info("Scheduling flat-fields...")
        await self._scheduler(Time.now())

        # log schedule
        log.info("Found schedule:")
        for sched in self._scheduler:
            log.info("- %s", sched)

        # total exposure time in ms
        self.exptime_done = 0

        # do flat fields
        item: SchedulerItem
        for item in self._scheduler:
            # do flat fields
            log.info("Performing flat-fields in %s %dx%d...", item.filter_name, *item.binning)
            if isinstance(flatfield, IBinning):
                await flatfield.set_binning(*item.binning)
            if isinstance(flatfield, IFilters):
                await flatfield.set_filter(item.filter_name)
            done, exp_time = await flatfield.flat_field(self._count)
            log.info("Finished flat-fields.")

            # increase exposure time
            self.exptime_done += exp_time

        # finished
        log.info("Finished all scheduled flat-fields.")


__all__ = ["SkyFlats"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class IWindow(Interface, metaclass=ABCMeta):
    """The camera supports windows, to be used together with :class:`~pyobs.interfaces.ICamera`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def get_full_frame(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns full size of CCD.

        Returns:
            Tuple with left, top, width, and height set.
        """
        ...

    @abstractmethod
    async def set_window(self, left: int, top: int, width: int, height: int, **kwargs: Any) -> None:
        """Set the camera window.

        Args:
            left: X offset of window.
            top: Y offset of window.
            width: Width of window.
            height: Height of window.

        Raises:
            ValueError: If window could not be set.
        """
        ...

    @abstractmethod
    async def get_window(self, **kwargs: Any) -> Tuple[int, int, int, int]:
        """Returns the camera window.

        Returns:
            Tuple with left, top, width, and height set.
        """
        ...


__all__ = ["IWindow"]



import logging
from typing import Any
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
import astropy.units as u

from pyobs.images import Image
from pyobs.images.meta import PixelOffsets, OnSkyDistance
from . import Offsets

log = logging.getLogger(__name__)


class CorrelationMaxCloseToBorderError(Exception):
    pass


class AstrometryOffsets(Offsets):
    """An offset-calculation method based on astrometry. Returns offset to real coordinates."""

    def __init__(self, **kwargs: Any):
        """Initializes new astrometry offsets.

        MUST run after an astrometry processor.
        """
        Offsets.__init__(self, **kwargs)

    async def __call__(self, image: Image) -> Image:
        """Processes an image and sets x/y pixel offset to reference in offset attribute.

        Args:
            image: Image to process.

        Returns:
            Original image.

        Raises:
            ValueError: If offset could not be found.
        """

        # copy image and get WCS
        # we make our life a little easier by only using the new WCS from astrometry
        img = image.copy()
        wcs = WCS(img.header)

        # get x/y coordinates from CRVAL1/2, i.e. from center with good WCS
        center = SkyCoord(img.header["CRVAL1"] * u.deg, img.header["CRVAL2"] * u.deg, frame="icrs")
        x_center, y_center = wcs.world_to_pixel(center)

        # get x/y coordinates from TEL-RA/-DEC, i.e. from where the telescope thought it's pointing
        tel = SkyCoord(img.header["TEL-RA"] * u.deg, img.header["TEL-DEC"] * u.deg, frame="icrs")
        x_tel, y_tel = wcs.world_to_pixel(tel)

        # calculate offsets as difference between both
        img.set_meta(PixelOffsets(x_tel - x_center, y_tel - y_center))
        img.set_meta(OnSkyDistance(center.separation(tel)))
        return img


__all__ = ["AstrometryOffsets"]



import asyncio
from abc import ABCMeta
from typing import Union, List, Dict, Tuple, Any, Optional
import logging
from astropy.coordinates import SkyCoord
import astropy.units as u

from pyobs.utils.time import Time
from pyobs.interfaces import IAutoGuiding, IFitsHeaderBefore
from pyobs.images import Image
from ._base import BasePointing
from ...interfaces import ITelescope

log = logging.getLogger(__name__)


class BaseGuiding(BasePointing, IAutoGuiding, IFitsHeaderBefore, metaclass=ABCMeta):
    """Base class for guiding modules."""

    __module__ = "pyobs.modules.pointing"

    def __init__(
        self,
        max_exposure_time: Optional[float] = None,
        min_interval: float = 0,
        max_interval: float = 600,
        separation_reset: Optional[float] = None,
        pid: bool = False,
        **kwargs: Any,
    ):
        """Initializes a new science frame auto guiding system.

        Args:
            max_exposure_time: Maximum exposure time in sec for images to analyse.
            min_interval: Minimum interval in sec between two images.
            max_interval: Maximum interval in sec between to consecutive images to guide.
            separation_reset: Min separation in arcsec between two consecutive images that triggers a reset.
            pid: Whether to use a PID for guiding.
        """
        BasePointing.__init__(self, **kwargs)

        # store
        self._enabled = False
        self._max_exposure_time = max_exposure_time
        self._min_interval = min_interval
        self._max_interval = max_interval
        self._separation_reset = separation_reset
        self._pid = pid
        self._loop_closed = False

        # headers of last and of reference image
        self._last_header = None
        self._ref_header = None

    async def start(self, **kwargs: Any) -> None:
        """Starts/resets auto-guiding."""
        log.info("Start auto-guiding...")
        await self._reset_guiding(enabled=True)

    async def stop(self, **kwargs: Any) -> None:
        """Stops auto-guiding."""
        log.info("Stopping autp-guiding...")
        await self._reset_guiding(enabled=False)

    async def is_running(self, **kwargs: Any) -> bool:
        """Whether auto-guiding is running.

        Returns:
            Auto-guiding is running.
        """
        return self._enabled

    async def get_fits_header_before(
        self, namespaces: Optional[List[str]] = None, **kwargs: Any
    ) -> Dict[str, Tuple[Any, str]]:
        """Returns FITS header for the current status of this module.

        Args:
            namespaces: If given, only return FITS headers for the given namespaces.

        Returns:
            Dictionary containing FITS headers.
        """

        # state
        state = "GUIDING_CLOSED_LOOP" if self._loop_closed else "GUIDING_OPEN_LOOP"

        # return header
        return {"AGSTATE": (state, "Autoguider state")}

    async def _reset_guiding(self, enabled: bool = True, image: Optional[Union[Image, None]] = None) -> None:
        """Reset guiding.

        Args:
            image: If given, new reference image.
        """
        self._enabled = enabled
        self._loop_closed = False
        self._ref_header = None if image is None else image.header
        self._last_header = None if image is None else image.header

        # reset offset
        self.reset_pipeline()
        if image is not None:
            # if image is given, process it
            loop = asyncio.get_running_loop()
            await self.run_pipeline(image)

    async def _process_image(self, image: Image) -> Optional[Image]:
        """Processes a single image and offsets telescope.

        Args:
            image: Image to process.
        """

        # not enabled?
        if not self._enabled:
            return None

        # we only accept OBJECT images
        if image.header["IMAGETYP"] != "object":
            return None

        # reference header?
        if self._ref_header is None:
            log.info("Setting new reference image...")
            await self._reset_guiding(image=image)
            return None

        # check RA/Dec in header and separation
        c1 = SkyCoord(ra=image.header["TEL-RA"] * u.deg, dec=image.header["TEL-DEC"] * u.deg, frame="icrs")
        c2 = SkyCoord(ra=self._ref_header["TEL-RA"] * u.deg, dec=self._ref_header["TEL-DEC"] * u.deg, frame="icrs")
        separation = c1.separation(c2).deg
        if self._separation_reset is not None and separation * 3600.0 > self._separation_reset:
            log.warning(
                'Nominal position of reference and new image differ by %.2f", resetting reference...',
                separation * 3600.0,
            )
            await self._reset_guiding(image=image)
            return None

        # check filter
        if (
            "FILTER" in image.header
            and "FILTER" in self._ref_header
            and image.header["FILTER"] != self._ref_header["FILTER"]
        ):
            log.warning("The filter has been changed since the last exposure, resetting reference...")
            await self._reset_guiding(image=image)
            return None

        # get time
        date_obs = Time(image.header["DATE-OBS"])

        # check times and focus
        if self._last_header is not None:
            # check times
            t0 = Time(self._last_header["DATE-OBS"])
            if (date_obs - t0).sec > self._max_interval:
                log.warning("Time between current and last image is too large, resetting reference...")
                await self._reset_guiding(image=image)
                return None
            if (date_obs - t0).sec < self._min_interval:
                log.warning("Time between current and last image is too small, ignoring image...")
                return None

            # check focus
            if "TEL-FOCU" in image.header:
                if abs(image.header["TEL-FOCU"] - self._last_header["TEL-FOCU"]) > 0.05:
                    log.warning("Focus difference between current and last image is too large, resetting reference...")
                    await self._reset_guiding(image=image)
                    return None

        # exposure time too large?
        if self._max_exposure_time is not None and image.header["EXPTIME"] > self._max_exposure_time:
            log.warning("Exposure time too large, skipping auto-guiding for now...")
            self._loop_closed = False
            return None

        # remember header
        self._last_header = image.header

        # get offset
        image = await self.run_pipeline(image)

        # get telescope
        try:
            telescope = await self.proxy(self._telescope, ITelescope)
        except ValueError:
            log.error("Given telescope does not exist or is not of correct type.")
            self._loop_closed = False
            return image

        # apply offsets
        if await self._apply(image, telescope, self.location):
            self._loop_closed = True
            log.info("Finished image.")
        else:
            log.info("Could not apply offsets.")
            self._loop_closed = False

        # return image, in case we added important data
        return image


__all__ = ["BaseGuiding"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .interface import Interface


class IImageGrabber(Interface, metaclass=ABCMeta):
    """The module can grab and return an image from whatever device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def grab_image(self, broadcast: bool = True, **kwargs: Any) -> str:
        """Grabs an image and returns reference.

        Args:
            broadcast: Broadcast existence of image.

        Returns:
            Name of image that was taken.

        Raises:
            GrabImageError: If there was a problem grabbing the image.
        """
        ...


__all__ = ["IImageGrabber"]



import logging
from typing import Union, Tuple, Optional, Any

from pyobs.interfaces import IRoof, ITelescope, IAcquisition, IAutoFocus
from pyobs.robotic import TaskRunner
from pyobs.robotic.scripts import Script
from pyobs.utils.enums import ImageType
from pyobs.utils.logger import DuplicateFilter
from pyobs.robotic import TaskSchedule, TaskArchive

log = logging.getLogger(__name__)

# logger for logging name of task
cannot_run_logger = logging.getLogger(__name__ + ":cannot_run")
cannot_run_logger.addFilter(DuplicateFilter())


class LcoAutoFocusScript(Script):
    """Auto focus script for LCO configs."""

    def __init__(
        self,
        roof: Optional[Union[str, IRoof]] = None,
        telescope: Optional[Union[str, ITelescope]] = None,
        acquisition: Optional[Union[str, IAcquisition]] = None,
        autofocus: Optional[Union[str, IAutoFocus]] = None,
        count: int = 5,
        step: float = 0.1,
        exptime: float = 2.0,
        **kwargs: Any,
    ):
        """Initialize a new LCO auto focus script.

        Args:
            roof: Roof to use
            telescope: Telescope to use
            acquisition: Acquisition to use
            autofocus: Autofocus to use
        """
        Script.__init__(self, **kwargs)

        # store
        self.roof = roof
        self.telescope = telescope
        self.acquisition = acquisition
        self.autofocus = autofocus
        self._count = count
        self._step = step
        self._exptime = exptime

        # get image type
        self.image_type = ImageType.OBJECT
        if self.configuration["type"] == "BIAS":
            self.image_type = ImageType.BIAS
        elif self.configuration["type"] == "DARK":
            self.image_type = ImageType.DARK

    async def _get_proxies(
        self,
    ) -> Tuple[Optional[IRoof], Optional[ITelescope], Optional[IAcquisition], Optional[IAutoFocus]]:
        """Get proxies for running the task

        Returns:
            Proxies for roof, telescope, acquisition, and autofocus

        Raises:
            ValueError: If could not get proxies for all modules
        """
        roof = await self.comm.safe_proxy(self.roof, IRoof)
        telescope = await self.comm.safe_proxy(self.telescope, ITelescope)
        acquisition = await self.comm.safe_proxy(self.acquisition, IAcquisition)
        autofocus = await self.comm.safe_proxy(self.autofocus, IAutoFocus)
        return roof, telescope, acquisition, autofocus

    async def can_run(self) -> bool:
        """Whether this config can currently run.

        Returns:
            True, if script can run now
        """

        # get proxies
        roof, telescope, acquisition, autofocus = await self._get_proxies()

        # need everything
        if roof is None or telescope is None or autofocus is None:
            cannot_run_logger.info("Cannot run task, no roof, telescope, or auto-focusser found.")
            return False

        # acquisition?
        if (
            "acquisition_config" in self.configuration
            and "mode" in self.configuration["acquisition_config"]
            and self.configuration["acquisition_config"]["mode"] == "ON"
            and acquisition is None
        ):
            cannot_run_logger.info("Cannot run task, no acquisition found.")
            return False

        # we need an open roof and a working telescope
        if not await roof.is_ready():
            cannot_run_logger.info("Cannot run task, roof not ready.")
            return False
        if not await telescope.is_ready():
            cannot_run_logger.info("Cannot run task, telescope not ready.")
            return False

        # seems alright
        return True

    async def run(
        self,
        task_runner: TaskRunner,
        task_schedule: Optional[TaskSchedule] = None,
        task_archive: Optional[TaskArchive] = None,
    ) -> None:
        """Run script.

        Raises:
            InterruptedError: If interrupted
        """

        # get proxies
        roof, telescope, acquisition, autofocus = await self._get_proxies()
        if telescope is None:
            raise ValueError("No telescope given.")

        # got a target?
        target = self.configuration["target"]
        log.info("Moving to target %s...", target["name"])
        await telescope.move_radec(target["ra"], target["dec"])

        # acquisition?
        if (
            "acquisition_config" in self.configuration
            and "mode" in self.configuration["acquisition_config"]
            and self.configuration["acquisition_config"]["mode"] == "ON"
        ):
            # TODO: unfortunately this never happens, since the LCO portal forces acquisition mode to OFF, see:
            # observation_portal/requestgroups/serializers.py:288 in portal code:
            # if data['type'] in ['LAMP_FLAT', 'ARC', 'AUTO_FOCUS', 'NRES_BIAS', 'NRES_DARK', 'BIAS', 'DARK', 'SCRIPT']:
            #     These types of observations should only ever be set to guiding mode OFF, but the acquisition modes for
            #     spectrographs won't necessarily have that mode. Force OFF here.
            #     data['acquisition_config']['mode'] = AcquisitionConfig.OFF

            # do acquisition
            if acquisition is None:
                raise ValueError("No acquisition given.")
            log.info("Performing acquisition...")
            await acquisition.acquire_target()

        # do auto focus
        if autofocus is None:
            raise ValueError("No autofocus given.")
        await autofocus.auto_focus(self._count, self._step, self._exptime)


__all__ = ["LcoAutoFocusScript"]



import asyncio
from urllib.parse import urljoin
import logging
from typing import List, Dict, Optional, Any, cast
import aiohttp as aiohttp
from astroplan import ObservingBlock
from astropy.time import TimeDelta
import astropy.units as u

from pyobs.robotic.task import Task
from pyobs.utils.time import Time
from pyobs.robotic.taskschedule import TaskSchedule
from .portal import Portal
from .task import LcoTask

log = logging.getLogger(__name__)


class LcoTaskSchedule(TaskSchedule):
    """Scheduler for using the LCO portal"""

    def __init__(
        self,
        url: str,
        site: str,
        token: str,
        enclosure: Optional[str] = None,
        telescope: Optional[str] = None,
        instrument: Optional[str] = None,
        period: int = 24,
        **kwargs: Any,
    ):
        """Creates a new LCO scheduler.

        Args:
            url: URL to portal
            site: Site filter for fetching requests
            token: Authorization token for portal
            portal_enclosure: Enclosure for new schedules.
            portal_telescope: Telescope for new schedules.
            portal_instrument: Instrument for new schedules.
            period: Period to schedule in hours
        """
        TaskSchedule.__init__(self, **kwargs)

        # portal
        self._portal = Portal(url, token)

        # store stuff
        self._url = url
        self._site = site
        self._enclosure = enclosure
        self._telescope = telescope
        self._instrument = instrument
        self._period = TimeDelta(period * u.hour)
        self.instruments: Dict[str, Any] = {}
        self._last_schedule_time: Optional[Time] = None

        # buffers in case of errors
        self._last_scheduled: Optional[Time] = None
        self._last_changed: Optional[Time] = None

        # header
        self._token = token
        self._header = {"Authorization": "Token " + token}

        # task list
        self._tasks: Dict[str, LcoTask] = {}

    async def open(self) -> None:
        """Open scheduler."""
        await TaskSchedule.open(self)

        # get stuff from portal
        await self._init_from_portal()

        # start update thread
        asyncio.create_task(self._update_schedule())

    async def _init_from_portal(self) -> None:
        """Initialize scheduler from portal."""

        # get instruments
        # don't catch exception, we want to fail, if something goes wrong here
        data = await self._portal.instruments()

        # and store
        self.instruments = {k.lower(): v for k, v in data.items()}

    async def last_scheduled(self) -> Optional[Time]:
        """Returns time of last scheduler run."""

        # try to update time
        try:
            # get data
            self._last_scheduled = await self._portal.last_scheduled()

        finally:
            # even in case of errors, return last time
            return self._last_scheduled

    async def _update_schedule(self) -> None:
        """Update thread."""
        while True:
            # do actual update
            try:
                await self.update_now()
            except:
                log.exception("An exception occurred.")

            # sleep a little
            await asyncio.sleep(10)

    async def update_now(self, force: bool = False) -> None:
        """Update list of requests.

        Args:
            force: Force update.
        """

        # remember now
        now = Time.now()

        # get time of last scheduler run and check, whether we need an update, which is not the case, if
        # - we updated before
        # - AND last update was after last schedule update
        # - AND last update is less then 1 min ago
        # - AND force is set to False
        last_scheduled = await self.last_scheduled()
        if (
            self._last_schedule_time is not None
            and (last_scheduled is None or self._last_schedule_time >= last_scheduled)
            and self._last_schedule_time > now - TimeDelta(1.0 * u.minute)
            and force is False
        ):
            # need no update
            return

        # need update!
        try:
            tasks = await self.get_schedule(
                end_after=now, start_before=now + TimeDelta(24 * u.hour), include_running=False
            )
        except TimeoutError:
            log.error("Request timed out")
            await asyncio.sleep(60)
            return
        except RuntimeError:
            log.warning("Could not fetch schedule.")
            return

        # any changes?
        if sorted(tasks) != sorted(self._tasks):
            log.info("Task list changed, found %d task(s) to run.", len(tasks))
            for task_id, task in tasks.items():
                log.info(f"  - {task.start} to {task.end}: {task.name} (#{task_id})")

        # update
        self._tasks = cast(Dict[str, LcoTask], tasks)

        # finished
        self._last_schedule_time = now

    async def get_schedule(self, start_before: Time, end_after: Time, include_running: bool = True) -> Dict[str, Task]:
        """Fetch schedule from portal.

        Args:
            start_before: Task must start before this time.
            end_after: Task must end after this time.
            include_running: Whether to include a currently running task.

        Returns:
            Dictionary with tasks.

        Raises:
            Timeout: If request timed out.
            RuntimeError: If something goes wrong.
        """

        # define states
        states = ["PENDING"]
        if include_running:
            states += ["IN_PROGRESS"]

        # get url and params
        url = urljoin(self._url, "/api/observations/")
        params = {
            "site": self._site,
            "end_after": end_after.isot,
            "start_before": start_before.isot,
            "state": states,
            "request_state": "PENDING",
            "limit": 1000,
        }

        # do request
        async with aiohttp.ClientSession() as session:
            async with session.get(url, headers=self._header, params=params, timeout=10) as response:
                if response.status != 200:
                    raise RuntimeError("Invalid response from portal.")
                data = await response.json()

                # get schedule
                schedules = data["results"]

                # create tasks
                tasks = {}
                for sched in schedules:
                    # parse start and end
                    sched["start"] = Time(sched["start"])
                    sched["end"] = Time(sched["end"])

                    # create task
                    task = self._create_task(LcoTask, config=sched)
                    tasks[sched["request"]["id"]] = task

                # finished
                return tasks

    def get_task(self, time: Time) -> Optional[LcoTask]:
        """Returns the active task at the given time.

        Args:
            time: Time to return task for.

        Returns:
            Task at the given time or None.
        """

        # loop all tasks
        for task in self._tasks.values():
            # running now?
            if task.start <= time < task.end and not task.is_finished():
                return task

        # nothing found
        return None

    async def send_update(self, status_id: int, status: Dict[str, Any]) -> None:
        """Send report to LCO portal

        Args:
            status_id: id of config status
            status: Status dictionary
        """

        log.info("Sending configuration status update to portal...")
        url = urljoin(self._url, "/api/configurationstatus/%d/" % status_id)

        # do request
        try:
            async with aiohttp.ClientSession() as session:
                async with session.patch(url, json=status, headers=self._header, timeout=10) as response:
                    if response.status != 200:
                        log.error("Could not update configuration status: %s", await response.text())

        except TimeoutError:
            log.error("Request timed out.")

    async def set_schedule(self, blocks: List[ObservingBlock], start_time: Time) -> None:
        """Update the list of scheduled blocks.

        Args:
            blocks: Scheduled blocks.
            start_time: Start time for schedule.
        """

        # create observations
        observations = self._create_observations(blocks)

        # cancel schedule
        await self._cancel_schedule(start_time)

        # send new schedule
        await self._submit_observations(observations)

    async def _cancel_schedule(self, now: Time) -> None:
        """Cancel future schedule."""

        # define parameters
        params = {
            "site": self._site,
            "enclosure": self._enclosure,
            "telescope": self._telescope,
            "start": now.isot,
            "end": (now + self._period).isot,
        }

        # url and headers
        url = urljoin(self._url, "/api/observations/cancel/")
        headers = {"Authorization": "Token " + self._token, "Content-Type": "application/json; charset=utf8"}

        # cancel schedule
        log.info("Deleting all scheduled tasks after %s...", now.isot)
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=params, headers=headers, timeout=10) as response:
                if response.status != 200:
                    log.error("Could not cancel schedule: %s", await response.text())

    def _create_observations(self, blocks: List[ObservingBlock]) -> List[Dict[str, Any]]:
        """Create observations from schedule.

        Args:
            blocks: List of scheduled blocks

        Returns:
            List with observations.
        """

        # loop blocks
        observations = []
        for block in blocks:
            # get request
            request = block.configuration["request"]

            # add observation
            observations.append(
                {
                    "site": self._site,
                    "enclosure": self._enclosure,
                    "telescope": self._telescope,
                    "start": block.start_time.isot,
                    "end": block.end_time.isot,
                    "request": request["id"],
                    "configuration_statuses": [
                        {
                            "configuration": request["configurations"][0]["id"],
                            "instrument_name": self._instrument,
                            "guide_camera_name": self._instrument,
                        }
                    ],
                }
            )

        # return list
        return observations

    async def _submit_observations(self, observations: List[Dict[str, Any]]) -> None:
        """Submit observations.

        Args:
            observations: List of observations to submit.
        """

        # nothing?
        if len(observations) == 0:
            return

        # url and headers
        url = urljoin(self._url, "/api/observations/")
        headers = {"Authorization": "Token " + self._token, "Content-Type": "application/json; charset=utf8"}

        # submit observations
        async with aiohttp.ClientSession() as session:
            async with session.post(url, json=observations, headers=headers, timeout=10) as response:
                if response.status != 201:
                    log.error("Could not submit observations: %s", await response.text())
                data = await response.json()

        # log
        log.info("%d observations created.", data["num_created"])

        # errors?
        if "errors" in data and len(data["errors"]) > 0:
            for err in data["errors"].values():
                log.warning("Error from portal: " + str(err["non_field_errors"]))


__all__ = ["LcoTaskSchedule"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, Any

from .interface import Interface


class IOffsetsRaDec(Interface, metaclass=ABCMeta):
    """The module supports RA/Dec offsets, usually combined with :class:`~pyobs.interfaces.ITelescope` and
    :class:`~pyobs.interfaces.IPointingRaDec`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_offsets_radec(self, dra: float, ddec: float, **kwargs: Any) -> None:
        """Move an RA/Dec offset.

        Args:
            dra: RA offset in degrees.
            ddec: Dec offset in degrees.

        Raises:
            MoveError: If telescope cannot be moved.
        """
        ...

    @abstractmethod
    async def get_offsets_radec(self, **kwargs: Any) -> Tuple[float, float]:
        """Get RA/Dec offset.

        Returns:
            Tuple with RA and Dec offsets.
        """
        ...


__all__ = ["IOffsetsRaDec"]



from abc import ABCMeta, abstractmethod
from typing import Tuple, List, Any

from .interface import Interface


class IBinning(Interface, metaclass=ABCMeta):
    """The camera supports binning, to be used together with :class:`~pyobs.interfaces.ICamera`."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_binning(self, x: int, y: int, **kwargs: Any) -> None:
        """Set the camera binning.

        Args:
            x: X binning.
            y: Y binning.

        Raises:
            ValueError: If binning could not be set.
        """
        ...

    @abstractmethod
    async def get_binning(self, **kwargs: Any) -> Tuple[int, int]:
        """Returns the camera binning.

        Returns:
            Tuple with x and y.
        """
        ...

    @abstractmethod
    async def list_binnings(self, **kwargs: Any) -> List[Tuple[int, int]]:
        """List available binnings.

        Returns:
            List of available binnings as (x, y) tuples.
        """
        ...


__all__ = ["IBinning"]



"""
A virtual file system (VFS) is a convenient way for allowing access to different resources with a  common interface.
It also makes it easy to change the actual place of storage without any changes to the code. In *pyobs*, a path into
the VFS looks like a normal path on a Unix-like system::

    /path/to/file.ext

To be more precise, a path is built like this::

    /<root>/<path>/<filename>

where ``<root>`` indicates, which VFS to use, ``<path>`` is the path within that VFS and `<filename>` is the actual
filename.

The available roots in the *pyobs* VFS are usually defined in the configuration file like this::

    vfs:
        root1:
            class: VfsClass1
        root2:
            class: VfsClass2

With this configuration, all filenames that start with ``/root1/`` (e.g. ``/root1/path/filename.ext``) are handled
by ``VfsClass1``, while all filenames starting with ``/root2/`` use ``VfsClass2``. Both defined classes can have
additional parameters, which should also be given in the configuration.

With this example, one can easily see the advantages of using a VFS:

    1. File access to all roots look similar, we always open a filename like ``/root/path/file.ext``.

    2. Changing the handling class for one of the roots changes the way we access a file. For instance, if
       a file was always stored locally, but now we want to change that to a location using a SSH location, this can
       easily be accomplished by simply changing the configuration.

    3. Another advantage that we have not mentioned before is, that the same roots can use different handling classes on
       different machines. See below for details.

Imagine a file ``/storage/file.ext`` that maps to a local file on machine A, i.e. ``storage`` uses a handling class
that simply changes the filename to a local filename. On machine B we can now define the same root ``storage``, but use
a different handling class that, e.g., accesses the file via SSH. Still, the filename would be the same on both
machines. So A could store a file as ``/storage/file.ext``, send the filename to B, which then can access the file
via the same filename.

Currently supported are these types of file access:
    - :class:`~pyobs.vfs.LocalFile`: Local file on the machine the module is running on.
    - :class:`~pyobs.vfs.HttpFile`: File on a HTTP server.
    - :class:`~pyobs.vfs.MemoryFile`: File in memory.
    - :class:`~pyobs.vfs.SSHFile`: File on different machine that is accessible via SSH.
    - :class:`~pyobs.vfs.TarFile`: Wrapper for a dynamically created TAR file. Can only be read from.
    - :class:`~pyobs.vfs.TempFile`: Temporary file that will be deleted after being closed.
    - :class:`~pyobs.vfs.ArchiveFile`: Wrapper for a file in the :ref:`pyobs-archive` image archive.

The base class for all of these classes is :class:`~pyobs.vfs.VFSFile`.

"""
__title__ = "Virtual File System"

from .vfs import VirtualFileSystem
from .file import VFSFile
from .localfile import LocalFile
from .httpfile import HttpFile
from .memfile import MemoryFile
from .smbfile import SMBFile
from .sshfile import SSHFile
from .tempfile import TempFile
from .archivefile import ArchiveFile


__all__ = [
    "VirtualFileSystem",
    "VFSFile",
    "LocalFile",
    "HttpFile",
    "MemoryFile",
    "SMBFile",
    "SSHFile",
    "TempFile",
    "ArchiveFile",
]



import inspect
import logging
from typing import Any, List, Type

from pyobs.comm import Comm
from pyobs.interfaces import Interface
from pyobs.utils.parallel import Future

log = logging.getLogger(__name__)


class DummyComm(Comm):
    """A dummy implementation of the Comm interface."""

    def __init__(self, *args: Any, **kwargs: Any):
        """Creates a new dummy comm."""
        Comm.__init__(self, *args, **kwargs)

    @property
    def clients(self) -> List[str]:
        """Always return zero clients."""
        return []

    def get_interfaces(self, client: str) -> List[Type[Interface]]:
        """No interfaces implemented."""
        return []

    def _supports_interface(self, client: str, interface: Type[Interface]) -> bool:
        """Interfaces are never supported."""
        return False

    def execute(self, client: str, method: str, signature: inspect.Signature, *args: Any) -> Future:
        """Always fake a successful execution of a method."""
        return Future(empty=True)

    @property
    def name(self) -> str:
        """Name of this client, which is unknown."""
        return "module"


__all__ = ["DummyComm"]



from abc import ABCMeta

from .IStartStop import IStartStop


class IAutonomous(IStartStop, metaclass=ABCMeta):
    """The module does some autonomous actions, mainly used for warnings to users."""

    __module__ = "pyobs.interfaces"
    pass


__all__ = ["IAutonomous"]



from abc import ABCMeta, abstractmethod
from typing import Any

from .IMotion import IMotion


class IFocuser(IMotion, metaclass=ABCMeta):
    """The module is a focusing device."""

    __module__ = "pyobs.interfaces"

    @abstractmethod
    async def set_focus(self, focus: float, **kwargs: Any) -> None:
        """Sets new focus.

        Args:
            focus: New focus value.

        Raises:
            MoveError: If telescope cannot be moved.
            InterruptedError: If movement was aborted.
        """
        ...

    @abstractmethod
    async def set_focus_offset(self, offset: float, **kwargs: Any) -> None:
        """Sets focus offset.

        Args:
            offset: New focus offset.

        Raises:
            ValueError: If given value is invalid.
            MoveError: If telescope cannot be moved.
        """
        ...

    @abstractmethod
    async def get_focus(self, **kwargs: Any) -> float:
        """Return current focus.

        Returns:
            Current focus.
        """
        ...

    @abstractmethod
    async def get_focus_offset(self, **kwargs: Any) -> float:
        """Return current focus offset.

        Returns:
            Current focus offset.
        """
        ...


__all__ = ["IFocuser"]
